import pstats

from yappi import YFuncStats  # type: ignore[attr-defined]

def clear_stats() -> None: ...
def set_clock_type(clock_type: str) -> None: ...
def start() -> None: ...
def stop() -> None: ...
def get_func_stats() -> YFuncStats: ...
def convert2pstats(stats: YFuncStats) -> pstats.Stats: ...



from functools import cache

from pydantic_settings import BaseSettings, SettingsConfigDict


class PostgresSettings(BaseSettings):
    host: str
    user: str
    password: str
    port: int
    db: str

    model_config = SettingsConfigDict(
        env_file=".env.db",
        env_file_encoding="utf-8",
        env_prefix="postgres_",
    )


@cache
def get_postgres_settings() -> PostgresSettings:
    return PostgresSettings()



from http import HTTPStatus
from typing import TYPE_CHECKING
from unittest.mock import ANY
from uuid import uuid4

from starlette.testclient import TestClient

from testing.util import get_random_string, get_resources_path
from yak_server.cli.database import initialize_database

if TYPE_CHECKING:
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_bets_by_groups(app_with_valid_jwt_config: "FastAPI", engine_for_test: "Engine") -> None:
    client = TestClient(app_with_valid_jwt_config)

    initialize_database(engine_for_test, get_resources_path("test_compute_points_v1"))

    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": get_random_string(6),
            "first_name": get_random_string(6),
            "last_name": get_random_string(6),
            "password": get_random_string(9),
        },
    )

    assert response_signup.status_code == HTTPStatus.CREATED

    access_token = response_signup.json()["result"]["access_token"]

    # Retrieve all groups
    groups_response = client.get(
        "/api/v1/groups",
        headers={"Authorization": f"Bearer {access_token}"},
    )

    assert groups_response.status_code == HTTPStatus.OK
    group_id = groups_response.json()["result"]["groups"][0]["id"]

    # Success case
    bets_by_valid_group = client.get(
        f"/api/v1/bets/groups/{group_id}",
        headers={"Authorization": f"Bearer {access_token}"},
    )

    assert bets_by_valid_group.json()["result"] == {
        "binary_bets": [
            {
                "id": ANY,
                "locked": False,
                "team1": None,
                "team2": None,
            },
        ],
        "group": {
            "code": "A",
            "description": "Groupe A",
            "id": ANY,
        },
        "phase": {
            "code": "GROUP",
            "description": "Phase de groupes",
            "id": ANY,
        },
        "score_bets": [
            {
                "id": ANY,
                "locked": False,
                "team1": {
                    "code": "FR",
                    "description": "France",
                    "flag": {"url": ANY},
                    "id": ANY,
                    "score": None,
                },
                "team2": {
                    "code": "IE",
                    "description": "Irlande",
                    "flag": {"url": ANY},
                    "id": ANY,
                    "score": None,
                },
            },
            {
                "id": ANY,
                "locked": False,
                "team1": {
                    "code": "FR",
                    "description": "France",
                    "flag": {"url": ANY},
                    "id": ANY,
                    "score": None,
                },
                "team2": {
                    "code": "IM",
                    "description": "Île de Man",
                    "flag": {"url": ANY},
                    "id": ANY,
                    "score": None,
                },
            },
            {
                "id": ANY,
                "locked": False,
                "team1": {
                    "code": "IE",
                    "description": "Irlande",
                    "flag": {"url": ANY},
                    "id": ANY,
                    "score": None,
                },
                "team2": {
                    "code": "IM",
                    "description": "Île de Man",
                    "flag": {"url": ANY},
                    "id": ANY,
                    "score": None,
                },
            },
        ],
    }

    assert len(bets_by_valid_group.json()["result"]["score_bets"]) == 3
    assert [
        (score_bet["team1"]["flag"]["url"], score_bet["team2"]["flag"]["url"])
        for score_bet in bets_by_valid_group.json()["result"]["score_bets"]
    ] == [
        ("https://fake-team-flag_france.com", "https://fake-team-flag_ireland.com"),
        ("https://fake-team-flag_france.com", "https://fake-team-flag_isle_of_man.com"),
        ("https://fake-team-flag_ireland.com", "https://fake-team-flag_isle_of_man.com"),
    ]

    # Error case : invalid group code
    invalid_group_code = uuid4()

    bets_by_invalid_group_code = client.get(
        f"/api/v1/bets/groups/{invalid_group_code}",
        headers={"Authorization": f"Bearer {access_token}"},
    )

    assert bets_by_invalid_group_code.json() == {
        "ok": False,
        "error_code": HTTPStatus.NOT_FOUND,
        "description": f"Group not found: {invalid_group_code}",
    }



from itertools import chain
from typing import TYPE_CHECKING

from fastapi import status
from pydantic import BaseModel
from sqlalchemy.orm import selectinload

from yak_server.database.models import BinaryBetModel, GroupModel, MatchModel, PhaseModel
from yak_server.helpers.group_position import get_group_rank_with_code

if TYPE_CHECKING:
    from sqlalchemy.orm import Session

    from yak_server.database.models import UserModel


class Team(BaseModel):
    rank: int
    group: str


class Versus(BaseModel):
    team1: Team
    team2: Team


class RuleComputeFinaleFromGroupRank(BaseModel):
    to_group: str
    from_phase: str
    versus: list[Versus]


def compute_finale_phase_from_group_rank(
    db: "Session",
    user: "UserModel",
    rule_config: RuleComputeFinaleFromGroupRank,
) -> tuple[int, str]:
    first_phase_group = db.query(GroupModel).filter_by(code=rule_config.to_group).first()

    if first_phase_group is None:
        return status.HTTP_404_NOT_FOUND, f"Group not found with code: {rule_config.to_group}"

    groups_result = {
        group.code: get_group_rank_with_code(db, user, group.id)
        for group in db
        .query(GroupModel)
        .join(GroupModel.phase)
        .where(
            PhaseModel.code == rule_config.from_phase,
        )
    }

    for index, match_config in enumerate(rule_config.versus, 1):
        if all(
            team.played == len(groups_result[match_config.team1.group]) - 1
            for team in chain(
                groups_result[match_config.team1.group],
                groups_result[match_config.team2.group],
            )
        ):
            team1 = groups_result[match_config.team1.group][match_config.team1.rank - 1].team
            team2 = groups_result[match_config.team2.group][match_config.team2.rank - 1].team

            binary_bet = (
                db
                .query(BinaryBetModel)
                .options(selectinload(BinaryBetModel.match))
                .join(BinaryBetModel.match)
                .where(
                    MatchModel.index == index,
                    MatchModel.user_id == user.id,
                    MatchModel.group_id == first_phase_group.id,
                )
                .first()
            )

            if binary_bet is not None:
                binary_bet.match.team1_id = team1.id
                binary_bet.match.team2_id = team2.id

            db.flush()

    db.commit()

    return status.HTTP_200_OK, ""



import json
from pathlib import Path
from typing import TYPE_CHECKING
from uuid import UUID

from sqlalchemy.dialects.postgresql import insert

from yak_server.database.models import (
    Base,
    BinaryBetModel,
    GroupModel,
    GroupPositionModel,
    MatchModel,
    MatchReferenceModel,
    PhaseModel,
    ScoreBetModel,
    TeamModel,
    UserModel,
)
from yak_server.database.session import build_local_session_maker

if TYPE_CHECKING:
    from sqlalchemy import Engine
    from sqlalchemy.orm import Session


class RecordDeletionInProductionError(Exception):
    def __init__(self) -> None:
        super().__init__("Trying to delete records in production using script. DO NOT DO IT.")


class TableDropInProductionError(Exception):
    def __init__(self) -> None:
        super().__init__("Trying to drop database tables in production using script. DO NOT DO IT.")


def create_database(engine: "Engine") -> None:
    Base.metadata.create_all(bind=engine)


class MissingPhaseDuringInitError(Exception):
    def __init__(self, phase_code: str) -> None:
        super().__init__(
            f"Error during database initialization: phase_code={phase_code} not found.",
        )


class MissingTeamDuringInitError(Exception):
    def __init__(self, team_code: str) -> None:
        super().__init__(f"Error during database initialization: team_code={team_code} not found.")


class MissingGroupDuringInitError(Exception):
    def __init__(self, group_code: str) -> None:
        super().__init__(
            f"Error during database initialization: group_code={group_code} not found.",
        )


def fetch_team_id(team_code: str | None, db: "Session") -> UUID | None:
    if team_code is None:
        return None

    team = db.query(TeamModel).filter_by(code=team_code).first()

    if team is None:
        raise MissingTeamDuringInitError(team_code)

    return team.id


def fetch_group_id(group_code: str, db: "Session") -> UUID:
    group = db.query(GroupModel).filter_by(code=group_code).first()

    if group is None:
        raise MissingGroupDuringInitError(group_code)

    return group.id


def initialize_database(engine: "Engine", data_folder: Path) -> None:
    local_session_maker = build_local_session_maker(engine)

    with local_session_maker() as db:
        phases = json.loads((data_folder / "phases.json").read_text(encoding="utf-8"))

        for phase in phases:
            stmt = insert(PhaseModel).values(**phase)
            db.execute(
                stmt.on_conflict_do_update(
                    index_elements=["code"],
                    set_={col: stmt.excluded[col] for col in phase if col != "code"},
                ),
            )

        db.flush()

        groups = json.loads((data_folder / "groups.json").read_text(encoding="utf-8"))

        for group in groups:
            phase_code = group.pop("phase_code")

            phase = db.query(PhaseModel).filter_by(code=phase_code).first()

            if phase is None:
                raise MissingPhaseDuringInitError(phase_code)

            group["phase_id"] = phase.id

        for group in groups:
            stmt = insert(GroupModel).values(**group)
            db.execute(
                stmt.on_conflict_do_update(
                    index_elements=["code"],
                    set_={col: stmt.excluded[col] for col in group if col != "code"},
                ),
            )

        db.flush()

        teams = json.loads((data_folder / "teams.json").read_text(encoding="utf-8"))

        for team in teams:
            team["flag_url"] = team["internal_flag_url"]

            stmt = insert(TeamModel).values(**team)
            db.execute(
                stmt.on_conflict_do_update(
                    index_elements=["code"],
                    set_={col: stmt.excluded[col] for col in team if col != "code"},
                ),
            )

            db.flush()

        matches = json.loads((data_folder / "matches.json").read_text(encoding="utf-8"))

        for match in matches:
            match["team1_id"] = fetch_team_id(match.pop("team1_code"), db)
            match["team2_id"] = fetch_team_id(match.pop("team2_code"), db)
            match["group_id"] = fetch_group_id(match.pop("group_code"), db)

        for match in matches:
            stmt = insert(MatchReferenceModel).values(**match)
            db.execute(
                stmt.on_conflict_do_update(
                    index_elements=["group_id", "index"],
                    set_={
                        col: stmt.excluded[col] for col in match if col not in {"group_id", "index"}
                    },
                ),
            )

        db.flush()

        db.commit()


def delete_database(engine: "Engine", *, debug: bool) -> None:
    if debug is False:
        raise RecordDeletionInProductionError

    local_session_maker = build_local_session_maker(engine)

    with local_session_maker() as db:
        db.query(GroupPositionModel).delete()
        db.query(ScoreBetModel).delete()
        db.query(BinaryBetModel).delete()
        db.query(MatchReferenceModel).delete()
        db.query(MatchModel).delete()
        db.query(UserModel).delete()
        db.query(GroupModel).delete()
        db.query(PhaseModel).delete()
        db.query(TeamModel).delete()
        db.commit()


def drop_database(engine: "Engine", *, debug: bool) -> None:
    if debug is False:
        raise TableDropInProductionError

    Base.metadata.drop_all(bind=engine)



"""Extend group code size

Revision ID: c2f508437a7e
Revises: d09424392322
Create Date: 2025-12-22 23:26:24.089512

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c2f508437a7e"
down_revision = "d09424392322"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "group",
        "code",
        existing_type=sa.VARCHAR(length=1),
        type_=sa.String(length=2),
        existing_nullable=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "group",
        "code",
        existing_type=sa.String(length=2),
        type_=sa.VARCHAR(length=1),
        existing_nullable=False,
    )
    # ### end Alembic commands ###



from enum import Enum
from typing import Protocol


class Lang(str, Enum):
    fr = "fr"
    en = "en"


class MultiLingualDescription(Protocol):
    description_fr: str
    description_en: str


def get_language_description(instance: MultiLingualDescription, lang: Lang) -> str:
    if lang == Lang.fr:
        return instance.description_fr

    return instance.description_en


DEFAULT_LANGUAGE = Lang.fr



from typing import TYPE_CHECKING

import click

from yak_server.database.models import Role
from yak_server.database.session import build_local_session_maker
from yak_server.helpers.authentication import NameAlreadyExistsError, signup_user

if TYPE_CHECKING:
    from sqlalchemy import Engine


def create_admin(password: str, engine: "Engine") -> None:
    local_session_maker = build_local_session_maker(engine)

    with local_session_maker() as db:
        try:
            _ = signup_user(
                db,
                name="admin",
                first_name="admin",
                last_name="admin",
                password=password,
                role=Role.ADMIN,
            )
        except NameAlreadyExistsError:
            click.echo("Admin already exists")



from datetime import datetime, timezone

from yak_server.database.models import UserModel

from .authentication import Permission, has_permission


def is_locked(user: UserModel, lock_datetime: datetime) -> bool:
    return not has_permission(user, Permission.ADMIN) and datetime.now(timezone.utc) > lock_datetime



from http import HTTPStatus
from typing import TYPE_CHECKING
from uuid import uuid4

from starlette.testclient import TestClient

from testing.util import get_random_string

if TYPE_CHECKING:
    from fastapi import FastAPI


def test_rule(app_with_empty_rules: "FastAPI") -> None:
    client = TestClient(app_with_empty_rules)

    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": get_random_string(6),
            "first_name": get_random_string(6),
            "last_name": get_random_string(6),
            "password": get_random_string(13),
        },
    )

    assert response_signup.status_code == HTTPStatus.CREATED

    invalid_rule_id = uuid4()

    response_execute_rule = client.post(
        f"/api/v1/rules/{invalid_rule_id}",
        headers={"Authorization": f"Bearer {response_signup.json()['result']['access_token']}"},
    )

    assert response_execute_rule.status_code == HTTPStatus.NOT_FOUND
    assert response_execute_rule.json() == {
        "ok": False,
        "error_code": HTTPStatus.NOT_FOUND,
        "description": f"Rule not found: {invalid_rule_id}",
    }



"""Add unique constraint on index for phase and group

Revision ID: a1b2c3d4e5f6
Revises: c2f508437a7e
Create Date: 2026-02-17 00:00:00.000000

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "a1b2c3d4e5f6"
down_revision = "c2f508437a7e"
branch_labels = None
depends_on = None


def upgrade():
    op.create_unique_constraint("uq_phase_index", "phase", ["index"])
    op.create_unique_constraint("uq_group_index", "group", ["index"])


def downgrade():
    op.drop_constraint("uq_group_index", "group", type_="unique")
    op.drop_constraint("uq_phase_index", "phase", type_="unique")



from fastapi import Response

from .settings import CookieSettings

ACCESS_TOKEN_COOKIE = "access_token"  # noqa: S105 - This is not a secret, just the name of the cookie.
REFRESH_TOKEN_COOKIE = "refresh_token"  # noqa: S105 - This is not a secret, just the name of the cookie.


def set_auth_cookies(
    response: Response,
    access_token: str,
    access_expires_in: int,
    refresh_token: str,
    refresh_expires_in: int,
    settings: CookieSettings,
) -> None:
    response.set_cookie(
        key=ACCESS_TOKEN_COOKIE,
        value=access_token,
        max_age=access_expires_in,
        path="/api",
        httponly=True,
        secure=settings.cookie_secure,
        samesite="lax",
        domain=settings.cookie_domain or None,
    )
    response.set_cookie(
        key=REFRESH_TOKEN_COOKIE,
        value=refresh_token,
        max_age=refresh_expires_in,
        path="/api/v1/users/refresh",
        httponly=True,
        secure=settings.cookie_secure,
        samesite="lax",
        domain=settings.cookie_domain or None,
    )


def clear_auth_cookies(response: Response, settings: CookieSettings) -> None:
    response.delete_cookie(
        key=ACCESS_TOKEN_COOKIE,
        path="/api",
        httponly=True,
        secure=settings.cookie_secure,
        samesite="lax",
        domain=settings.cookie_domain or None,
    )
    response.delete_cookie(
        key=REFRESH_TOKEN_COOKIE,
        path="/api/v1/users/refresh",
        httponly=True,
        secure=settings.cookie_secure,
        samesite="lax",
        domain=settings.cookie_domain or None,
    )



from collections.abc import Callable
from dataclasses import dataclass
from uuid import UUID

from pydantic import BaseModel
from sqlalchemy.orm import Session

from yak_server.database.models import UserModel

from .compute_final_from_rank import (
    RuleComputeFinaleFromGroupRank,
    compute_finale_phase_from_group_rank,
)
from .compute_points import RuleComputePoints
from .compute_points import compute_points as compute_points_func


class Rules(BaseModel):
    compute_finale_phase_from_group_rank: RuleComputeFinaleFromGroupRank | None = None
    compute_points: RuleComputePoints | None = None


@dataclass(frozen=True, kw_only=True)
class RuleMetadata:
    function: (
        Callable[[Session, UserModel, RuleComputeFinaleFromGroupRank], tuple[int, str]]
        | Callable[[Session, UserModel, RuleComputePoints], tuple[int, str]]
    )
    attribute: str
    required_admin: bool = False


RULE_MAPPING = {
    UUID("492345de-8d4a-45b6-8b94-d219f2b0c3e9"): RuleMetadata(
        function=compute_finale_phase_from_group_rank,
        attribute="compute_finale_phase_from_group_rank",
    ),
    UUID("62d46542-8cf1-4a3b-af77-a5086f10ac59"): RuleMetadata(
        function=compute_points_func,
        attribute="compute_points",
        required_admin=True,
    ),
}

# Static check to make sure RULE_MAPPING and Rules attributes are same
rule_mapping_attributes = {rule_metadata.attribute for rule_metadata in RULE_MAPPING.values()}
rule_class_attributes = set(Rules.model_fields)

if rule_mapping_attributes != rule_class_attributes:  # pragma: no cover
    msg = (
        "RULE_MAPPING doesn't have the same attributes as Rules"
        f" class:\n{rule_mapping_attributes}\n{rule_class_attributes}"
    )
    raise ValueError(msg)



from collections.abc import Generator
from http import HTTPStatus
from typing import TYPE_CHECKING

import pytest
from click.testing import CliRunner
from starlette.testclient import TestClient

from testing.mock import MockSettings
from testing.util import UserData, get_random_string, get_resources_path, patch_score_bets
from yak_server.cli import app as cli_app
from yak_server.cli.admin import create_admin
from yak_server.cli.database import initialize_database
from yak_server.helpers.rules import Rules
from yak_server.helpers.rules.compute_final_from_rank import (
    RuleComputeFinaleFromGroupRank,
    Team,
    Versus,
)
from yak_server.helpers.rules.compute_points import RuleComputePoints
from yak_server.helpers.settings import get_settings

if TYPE_CHECKING:
    from fastapi import FastAPI
    from sqlalchemy import Engine


def put_finale_phase(client: TestClient, access_token: str, *, is_one_won: bool | None) -> None:
    response_post_finale_phase_bets_admin = client.post(
        "/api/v1/rules/492345de-8d4a-45b6-8b94-d219f2b0c3e9",
        headers={"Authorization": f"Bearer {access_token}"},
    )

    assert response_post_finale_phase_bets_admin.json() == {"ok": True, "result": ""}
    assert response_post_finale_phase_bets_admin.status_code == HTTPStatus.OK

    response_get_finale_phase = client.get(
        "/api/v1/bets/phases/FINAL",
        headers={"Authorization": f"Bearer {access_token}"},
    )

    assert response_get_finale_phase.status_code == HTTPStatus.OK

    response_patch_finale_phase = client.patch(
        f"/api/v1/binary_bets/{response_get_finale_phase.json()['result']['binary_bets'][0]['id']}",
        json={"is_one_won": is_one_won},
        headers={"Authorization": f"Bearer {access_token}"},
    )

    assert response_patch_finale_phase.status_code == HTTPStatus.OK


@pytest.fixture
def app_and_rules_for_compute_points(
    app_with_valid_jwt_config: "FastAPI",
) -> Generator[tuple["FastAPI", Rules], None, None]:
    rules = Rules(
        compute_finale_phase_from_group_rank=RuleComputeFinaleFromGroupRank(
            to_group="1",
            from_phase="GROUP",
            versus=[Versus(team1=Team(rank=1, group="A"), team2=Team(rank=2, group="A"))],
        ),
        compute_points=RuleComputePoints(
            base_correct_result=1,
            multiplying_factor_correct_result=2,
            base_correct_score=3,
            multiplying_factor_correct_score=7,
            team_qualified=10,
            first_team_qualified=20,
        ),
    )

    app_with_valid_jwt_config.dependency_overrides[get_settings] = MockSettings(rules=rules)

    yield app_with_valid_jwt_config, rules

    app_with_valid_jwt_config.dependency_overrides.clear()


def test_compute_points(
    app_and_rules_for_compute_points: tuple["FastAPI", Rules],
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    app, rules = app_and_rules_for_compute_points

    client = TestClient(app)

    initialize_database(engine_for_test, get_resources_path("test_compute_points_v1"))

    # Signup admin
    password = get_random_string(15)

    admin = UserData(
        first_name="admin",
        last_name="admin",
        name="admin",
        scores=[(1, 2), (5, 1), (5, 5)],
    )

    create_admin(password, engine_for_test)

    response_login_admin = client.post(
        "/api/v1/users/login",
        json={
            "name": admin.name,
            "password": password,
        },
    )

    assert response_login_admin.status_code == HTTPStatus.CREATED

    admin.access_token = response_login_admin.json()["result"]["access_token"]

    # Patch admin scores
    patch_score_bets(client, admin.access_token, admin.scores)

    # Signup 3 players and patch their scores
    users_data = [
        UserData(
            first_name=get_random_string(15),
            last_name=get_random_string(15),
            name="user1",
            scores=[(2, 2), (5, 1), (5, 5)],
        ),
        UserData(
            first_name=get_random_string(15),
            last_name=get_random_string(15),
            name="user2",
            scores=[(2, 0), (2, 0), (1, 4)],
        ),
        UserData(
            first_name=get_random_string(15),
            last_name=get_random_string(15),
            name="user3",
            scores=[(0, 2), (2, 0), (1, 1)],
        ),
    ]

    for user_data in users_data:
        response_signup = client.post(
            "/api/v1/users/signup",
            json={
                "name": user_data.name,
                "first_name": user_data.first_name,
                "last_name": user_data.last_name,
                "password": get_random_string(18),
            },
        )

        assert response_signup.status_code == HTTPStatus.CREATED

        user_data.access_token = response_signup.json()["result"]["access_token"]

        patch_score_bets(client, user_data.access_token, user_data.scores)

    # Success case : compute_points call
    response_compute_points = client.post(
        "/api/v1/rules/62d46542-8cf1-4a3b-af77-a5086f10ac59",
        headers={"Authorization": f"Bearer {admin.access_token}"},
    )

    assert response_compute_points.status_code == HTTPStatus.OK

    # Error case : check unauthorized admin access to compute_points
    response_compute_points_unauthorized = client.post(
        "/api/v1/rules/62d46542-8cf1-4a3b-af77-a5086f10ac59",
        headers={"Authorization": f"Bearer {users_data[0].access_token}"},
    )

    assert response_compute_points_unauthorized.status_code == HTTPStatus.UNAUTHORIZED
    assert response_compute_points_unauthorized.json() == {
        "ok": False,
        "error_code": HTTPStatus.UNAUTHORIZED,
        "description": "Unauthorized access to admin API",
    }

    # Success case : Check score board call
    score_board_response = client.get(
        "/api/v1/score_board",
        headers={"Authorization": f"Bearer {users_data[0].access_token}"},
    )

    assert score_board_response.json()["result"] == [
        {
            "rank": 1,
            "first_name": users_data[2].first_name,
            "full_name": f"{users_data[2].first_name} {users_data[2].last_name}",
            "last_name": users_data[2].last_name,
            "number_final_guess": 0,
            "number_first_qualified_guess": 1,
            "number_match_guess": 3,
            "number_qualified_teams_guess": 2,
            "number_quarter_final_guess": 0,
            "number_score_guess": 0,
            "number_semi_final_guess": 0,
            "number_winner_guess": 0,
            "points": 46.0,
        },
        {
            "rank": 2,
            "first_name": users_data[0].first_name,
            "full_name": f"{users_data[0].first_name} {users_data[0].last_name}",
            "last_name": users_data[0].last_name,
            "number_final_guess": 0,
            "number_first_qualified_guess": 0,
            "number_match_guess": 2,
            "number_qualified_teams_guess": 2,
            "number_quarter_final_guess": 0,
            "number_score_guess": 2,
            "number_semi_final_guess": 0,
            "number_winner_guess": 0,
            "points": 43.0,
        },
        {
            "rank": 3,
            "first_name": users_data[1].first_name,
            "full_name": f"{users_data[1].first_name} {users_data[1].last_name}",
            "last_name": users_data[1].last_name,
            "number_final_guess": 0,
            "number_first_qualified_guess": 0,
            "number_match_guess": 1,
            "number_qualified_teams_guess": 1,
            "number_quarter_final_guess": 0,
            "number_score_guess": 0,
            "number_semi_final_guess": 0,
            "number_winner_guess": 0,
            "points": 11.0,
        },
    ]

    # Push finale phase bets
    put_finale_phase(client, admin.access_token, is_one_won=True)
    put_finale_phase(client, users_data[0].access_token, is_one_won=False)
    put_finale_phase(client, users_data[1].access_token, is_one_won=True)
    put_finale_phase(client, users_data[2].access_token, is_one_won=False)

    # Compute points again with cli
    monkeypatch.setattr("yak_server.cli.score_board.get_settings", MockSettings(rules=rules))

    runner = CliRunner()

    result = runner.invoke(cli_app, ["db", "score-board"])

    assert result.exit_code == 0

    # Check score board after finale phase bets patch
    score_board_response_after_finale_phase = client.get(
        "/api/v1/score_board",
        headers={"Authorization": f"Bearer {users_data[0].access_token}"},
    )

    assert score_board_response_after_finale_phase.json()["result"] == [
        {
            "rank": 1,
            "first_name": users_data[0].first_name,
            "full_name": f"{users_data[0].first_name} {users_data[0].last_name}",
            "last_name": users_data[0].last_name,
            "number_final_guess": 2,
            "number_first_qualified_guess": 0,
            "number_match_guess": 2,
            "number_qualified_teams_guess": 2,
            "number_quarter_final_guess": 0,
            "number_score_guess": 2,
            "number_semi_final_guess": 0,
            "number_winner_guess": 1,
            "points": 483.0,
        },
        {
            "rank": 2,
            "first_name": users_data[2].first_name,
            "full_name": f"{users_data[2].first_name} {users_data[2].last_name}",
            "last_name": users_data[2].last_name,
            "number_final_guess": 2,
            "number_first_qualified_guess": 1,
            "number_match_guess": 3,
            "number_qualified_teams_guess": 2,
            "number_quarter_final_guess": 0,
            "number_score_guess": 0,
            "number_semi_final_guess": 0,
            "number_winner_guess": 0,
            "points": 286.0,
        },
        {
            "rank": 3,
            "first_name": users_data[1].first_name,
            "full_name": f"{users_data[1].first_name} {users_data[1].last_name}",
            "last_name": users_data[1].last_name,
            "number_final_guess": 1,
            "number_first_qualified_guess": 0,
            "number_match_guess": 1,
            "number_qualified_teams_guess": 1,
            "number_quarter_final_guess": 0,
            "number_score_guess": 0,
            "number_semi_final_guess": 0,
            "number_winner_guess": 0,
            "points": 131.0,
        },
    ]

    # Success case : check user GET /results call
    get_results_response = client.get(
        "/api/v1/results",
        headers={"Authorization": f"Bearer {users_data[0].access_token}"},
    )

    assert get_results_response.json()["result"] == {
        "rank": 1,
        "first_name": users_data[0].first_name,
        "last_name": users_data[0].last_name,
        "full_name": f"{users_data[0].first_name} {users_data[0].last_name}",
        "number_final_guess": 2,
        "number_first_qualified_guess": 0,
        "number_match_guess": 2,
        "number_qualified_teams_guess": 2,
        "number_quarter_final_guess": 0,
        "number_score_guess": 2,
        "number_semi_final_guess": 0,
        "number_winner_guess": 1,
        "points": 483.0,
    }

    # Error case : check not result for admin user
    get_results_response_admin = client.get(
        "/api/v1/results",
        headers={"Authorization": f"Bearer {admin.access_token}"},
    )

    assert get_results_response_admin.json() == {
        "ok": True,
        "result": {
            "rank": 0,
            "first_name": admin.first_name,
            "last_name": admin.last_name,
            "full_name": f"{admin.first_name} {admin.last_name}",
            "number_final_guess": 0,
            "number_first_qualified_guess": 0,
            "number_match_guess": 0,
            "number_qualified_teams_guess": 0,
            "number_quarter_final_guess": 0,
            "number_score_guess": 0,
            "number_semi_final_guess": 0,
            "number_winner_guess": 0,
            "points": 0.0,
        },
    }



import psycopg
from sqlalchemy import URL, Engine, create_engine
from sqlalchemy.orm import Session, sessionmaker

from .settings import get_postgres_settings


def compute_database_uri(
    client: str, host: str, user: str, password: str, port: int, db: str
) -> URL:
    return URL.create(
        f"postgresql+{client}",
        username=user,
        password=password,
        host=host,
        port=port,
        database=db,
    )


def build_engine() -> Engine:
    postgres_settings = get_postgres_settings()

    database_url = compute_database_uri(
        psycopg.__name__,
        postgres_settings.host,
        postgres_settings.user,
        postgres_settings.password,
        postgres_settings.port,
        postgres_settings.db,
    )

    return create_engine(database_url, pool_recycle=7200, pool_pre_ping=True)


def build_local_session_maker(engine: Engine) -> sessionmaker[Session]:
    return sessionmaker(autocommit=False, autoflush=False, bind=engine)



import secrets
import string
from http import HTTPStatus
from typing import TYPE_CHECKING

import pytest
from starlette.testclient import TestClient

from testing.util import get_random_string
from yak_server.helpers.password_validator import (
    NoDigitError,
    NoLowerCaseError,
    NoUpperCaseError,
    SpaceError,
    TooShortError,
    validate_password,
)

if TYPE_CHECKING:
    from fastapi import FastAPI


def test_valid_password() -> None:
    validate_password(get_random_string(18))


def test_password_too_short() -> None:
    with pytest.raises(TooShortError) as exception:
        validate_password(get_random_string(6))

    assert str(exception.value) == "Password is too short. Minimum size is 8."


def test_password_without_upper_case_letter() -> None:
    with pytest.raises(NoUpperCaseError) as exception:
        validate_password(
            secrets.choice(string.ascii_lowercase)
            + secrets.choice(string.digits)
            + "".join(secrets.choice(string.ascii_lowercase + string.digits) for _ in range(10)),
        )

    assert str(exception.value) == "At least one upper-case letter expected."


def test_password_without_lower_case_letter() -> None:
    with pytest.raises(NoLowerCaseError) as exception:
        validate_password(
            secrets.choice(string.ascii_uppercase)
            + secrets.choice(string.digits)
            + "".join(secrets.choice(string.ascii_uppercase + string.digits) for _ in range(10)),
        )

    assert str(exception.value) == "At least one lower-case letter expected."


def test_password_without_digits() -> None:
    with pytest.raises(NoDigitError) as exception:
        validate_password(
            secrets.choice(string.ascii_uppercase)
            + secrets.choice(string.ascii_lowercase)
            + "".join(secrets.choice(string.ascii_letters) for _ in range(10)),
        )

    assert str(exception.value) == "At least one digit expected."


def test_password_with_spaces() -> None:
    with pytest.raises(SpaceError) as exception:
        validate_password(
            secrets.choice(string.ascii_uppercase)
            + secrets.choice(string.ascii_lowercase)
            + secrets.choice(string.digits)
            + "".join(secrets.choice(string.ascii_letters + string.digits) for _ in range(10))
            + " " * 10,
        )

    assert str(exception.value) == "Password must not contain spaces."


def test_password_requirements_v1(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config)

    response = client.get("/api/v1/users/signup/password_requirements")

    assert response.status_code == HTTPStatus.OK
    assert response.json() == {
        "ok": True,
        "result": {
            "minimum_length": 8,
            "uppercase": True,
            "lowercase": True,
            "digit": True,
            "no_space": True,
        },
    }



from functools import partial
from typing import Annotated

from fastapi import Depends, Request
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from jwt import ExpiredSignatureError, PyJWTError
from sqlalchemy.orm import Session

from yak_server.database.models import UserModel
from yak_server.helpers.authentication import Permission, decode_bearer_token, has_permission
from yak_server.helpers.cookies import ACCESS_TOKEN_COOKIE, REFRESH_TOKEN_COOKIE
from yak_server.helpers.database import get_db
from yak_server.helpers.settings import AuthenticationSettings, get_authentication_settings
from yak_server.v1.models.users import RefreshIn

from .errors import ExpiredToken, InvalidToken, UnauthorizedAccessToAdminAPI, UserNotFound

security = HTTPBearer(auto_error=False)


def user_from_token(db: Session, secret_key: str, access_token: str) -> UserModel:
    try:
        data = decode_bearer_token(access_token, secret_key)
    except ExpiredSignatureError as exc:
        raise ExpiredToken from exc
    except PyJWTError as exc:
        raise InvalidToken from exc

    user = db.query(UserModel).filter_by(id=data["sub"]).first()
    if not user:
        raise UserNotFound(data["sub"])

    return user


def _extract_token(
    request: Request,
    access_token: HTTPAuthorizationCredentials | None,
) -> str | None:
    if access_token is not None:
        return access_token.credentials

    return request.cookies.get(ACCESS_TOKEN_COOKIE)


def require_permission(
    required_permission: Permission,
    request: Request,
    access_token: Annotated[HTTPAuthorizationCredentials | None, Depends(security)],
    db: Annotated[Session, Depends(get_db)],
    auth_settings: Annotated[AuthenticationSettings, Depends(get_authentication_settings)],
) -> UserModel:
    token = _extract_token(request, access_token)

    if token is None:
        raise InvalidToken

    user = user_from_token(db, auth_settings.jwt_secret_key, token)

    if not has_permission(user, required_permission):
        raise UnauthorizedAccessToAdminAPI

    return user


require_user = partial(require_permission, Permission.USER)
require_admin = partial(require_permission, Permission.ADMIN)


def require_refresh_token(request: Request, refresh_in: RefreshIn | None = None) -> str:
    if refresh_in and refresh_in.refresh_token:
        return refresh_in.refresh_token

    token = request.cookies.get(REFRESH_TOKEN_COOKIE)

    if token is None:
        raise InvalidToken

    return token



"""Make some columns not nullable

Revision ID: 1719e9432d78
Revises: 6e28ca2ef0c7
Create Date: 2023-09-04 00:05:23.930630

"""

from alembic import op
from sqlalchemy.dialects import mysql

# revision identifiers, used by Alembic.
revision = "1719e9432d78"
down_revision = "6e28ca2ef0c7"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("team", "flag_url", existing_type=mysql.VARCHAR(length=100), nullable=False)
    op.alter_column(
        "team",
        "internal_flag_url",
        existing_type=mysql.VARCHAR(length=300),
        nullable=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "team",
        "internal_flag_url",
        existing_type=mysql.VARCHAR(length=300),
        nullable=True,
    )
    op.alter_column("team", "flag_url", existing_type=mysql.VARCHAR(length=100), nullable=True)
    # ### end Alembic commands ###



"""Add group position table.

Revision ID: 9f8e020eeced
Revises: f33dc66104ad
Create Date: 2023-03-02 23:40:21.051774

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "9f8e020eeced"
down_revision = "f33dc66104ad"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "group_position",
        sa.Column("id", sa.String(length=100), nullable=False),
        sa.Column("played", sa.Integer(), nullable=False),
        sa.Column("won", sa.Integer(), nullable=False),
        sa.Column("drawn", sa.Integer(), nullable=False),
        sa.Column("lost", sa.Integer(), nullable=False),
        sa.Column("goals_for", sa.Integer(), nullable=False),
        sa.Column("goals_against", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.String(length=100), nullable=False),
        sa.Column("team_id", sa.String(length=100), nullable=False),
        sa.Column("group_id", sa.String(length=100), nullable=False),
        sa.ForeignKeyConstraint(
            ["group_id"],
            ["group.id"],
        ),
        sa.ForeignKeyConstraint(
            ["team_id"],
            ["team.id"],
        ),
        sa.ForeignKeyConstraint(["user_id"], ["user.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("group_position")
    # ### end Alembic commands ###



"""Add refresh token table

Revision ID: e5f9dac72943
Revises: eb38661b5a0d
Create Date: 2025-07-20 15:39:15.144788

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "e5f9dac72943"
down_revision = "eb38661b5a0d"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "refresh_token",
        sa.Column("id", sa.UUID(), nullable=False),
        sa.Column("user_id", sa.UUID(), nullable=False),
        sa.Column("expiration", postgresql.TIMESTAMP(timezone=True), nullable=False),
        sa.ForeignKeyConstraint(["user_id"], ["user.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("refresh_token")
    # ### end Alembic commands ###



import time
from http import HTTPStatus
from typing import TYPE_CHECKING

from starlette.testclient import TestClient

from testing.util import get_random_string

if TYPE_CHECKING:
    from fastapi import FastAPI


def test_refresh_after_signup(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config)

    response = client.post(
        "/api/v1/users/signup",
        json={
            "name": get_random_string(10),
            "first_name": get_random_string(10),
            "last_name": get_random_string(10),
            "password": get_random_string(150),
        },
    )

    assert response.status_code == HTTPStatus.CREATED

    refresh_token = response.json()["result"]["refresh_token"]

    # Try to refresh the token
    response = client.post("/api/v1/users/refresh", json={"refresh_token": refresh_token})
    assert response.status_code == HTTPStatus.CREATED

    new_access_token = response.json()["result"]["access_token"]

    response = client.get(
        "/api/v1/bets",
        headers={"Authorization": f"Bearer {new_access_token}"},
    )

    assert response.status_code == HTTPStatus.OK


def test_refresh_after_login(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config)

    name = get_random_string(10)
    password = get_random_string(150)

    response = client.post(
        "/api/v1/users/signup",
        json={
            "name": name,
            "first_name": get_random_string(10),
            "last_name": get_random_string(10),
            "password": password,
        },
    )

    assert response.status_code == HTTPStatus.CREATED

    refresh_token = response.json()["result"]["refresh_token"]

    response = client.post(
        "/api/v1/users/login",
        json={
            "name": name,
            "password": password,
        },
    )

    assert response.status_code == HTTPStatus.CREATED

    # Try to refresh the token
    response = client.post("/api/v1/users/refresh", json={"refresh_token": refresh_token})
    assert response.status_code == HTTPStatus.CREATED

    new_access_token = response.json()["result"]["access_token"]

    response = client.get(
        "/api/v1/bets",
        headers={"Authorization": f"Bearer {new_access_token}"},
    )

    assert response.status_code == HTTPStatus.OK


def test_refresh_using_cookie(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config)

    response = client.post(
        "/api/v1/users/signup",
        json={
            "name": get_random_string(10),
            "first_name": get_random_string(10),
            "last_name": get_random_string(10),
            "password": get_random_string(150),
        },
    )

    assert response.status_code == HTTPStatus.CREATED

    # Call refresh without a body — the refresh token cookie set by signup should be used
    response = client.post("/api/v1/users/refresh")
    assert response.status_code == HTTPStatus.CREATED

    new_access_token = response.json()["result"]["access_token"]

    response = client.get(
        "/api/v1/bets",
        headers={"Authorization": f"Bearer {new_access_token}"},
    )

    assert response.status_code == HTTPStatus.OK


def test_refresh_without_token(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config, cookies={})

    response = client.post("/api/v1/users/refresh")

    assert response.status_code == HTTPStatus.UNAUTHORIZED
    assert response.json() == {
        "ok": False,
        "error_code": HTTPStatus.UNAUTHORIZED,
        "description": "Invalid access token, authentication required",
    }


def test_refresh_token_expired(app_with_null_jwt_refresh_expiration_time: "FastAPI") -> None:
    client = TestClient(app_with_null_jwt_refresh_expiration_time)

    # Login to get the refresh token
    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": get_random_string(10),
            "first_name": get_random_string(10),
            "last_name": get_random_string(10),
            "password": get_random_string(150),
        },
    )

    assert response_signup.status_code == HTTPStatus.CREATED

    refresh_token = response_signup.json()["result"]["refresh_token"]

    time.sleep(5)  # Wait for the refresh token to expire

    # Call the refresh endpoint with expired token:
    response = client.post("/api/v1/users/refresh", json={"refresh_token": refresh_token})

    assert response.status_code == HTTPStatus.UNAUTHORIZED
    assert response.json() == {
        "ok": False,
        "error_code": HTTPStatus.UNAUTHORIZED,
        "description": "Expired access token, re-authentication required",
    }


def test_refresh_token_cannot_access_protected_resources(
    app_with_valid_jwt_config: "FastAPI",
) -> None:
    """Test that refresh tokens cannot be used to access protected endpoints."""
    client = TestClient(app_with_valid_jwt_config)

    # Signup to get both tokens
    response = client.post(
        "/api/v1/users/signup",
        json={
            "name": get_random_string(10),
            "first_name": get_random_string(10),
            "last_name": get_random_string(10),
            "password": get_random_string(150),
        },
    )

    assert response.status_code == HTTPStatus.CREATED

    refresh_token = response.json()["result"]["refresh_token"]
    access_token = response.json()["result"]["access_token"]

    # Verify access token works for protected resources
    response_with_access_token = client.get(
        "/api/v1/bets",
        headers={"Authorization": f"Bearer {access_token}"},
    )
    assert response_with_access_token.status_code == HTTPStatus.OK

    # Try to access protected resource with refresh token - should fail
    response_with_refresh_token = client.get(
        "/api/v1/bets",
        headers={"Authorization": f"Bearer {refresh_token}"},
    )
    assert response_with_refresh_token.status_code == HTTPStatus.UNAUTHORIZED
    assert response_with_refresh_token.json() == {
        "ok": False,
        "error_code": HTTPStatus.UNAUTHORIZED,
        "description": "Invalid access token, authentication required",
    }

    # Test with another protected endpoint to ensure consistency
    response_current_user_with_refresh = client.get(
        "/api/v1/users/current",
        headers={"Authorization": f"Bearer {refresh_token}"},
    )
    assert response_current_user_with_refresh.status_code == HTTPStatus.UNAUTHORIZED
    assert response_current_user_with_refresh.json() == {
        "ok": False,
        "error_code": HTTPStatus.UNAUTHORIZED,
        "description": "Invalid access token, authentication required",
    }



from typing import TYPE_CHECKING, Annotated
from uuid import UUID

from fastapi import APIRouter, Depends, status
from sqlalchemy import func
from sqlalchemy.orm import Session

from yak_server.database.models import Role, UserModel
from yak_server.helpers.database import get_db
from yak_server.v1.helpers.auth import require_user
from yak_server.v1.models.generic import ErrorOut, GenericOut, ValidationErrorOut
from yak_server.v1.models.results import UserResult

if TYPE_CHECKING:
    from collections.abc import Sequence


router = APIRouter(tags=["results"])


@router.get(
    "/score_board",
    responses={
        status.HTTP_401_UNAUTHORIZED: {"model": ErrorOut},
        status.HTTP_422_UNPROCESSABLE_CONTENT: {"model": ValidationErrorOut},
    },
)
def retrieve_score_board(
    _: Annotated[UserModel, Depends(require_user)],
    db: Annotated[Session, Depends(get_db)],
) -> GenericOut[list[UserResult]]:
    return GenericOut(
        result=[
            UserResult.from_instance(user, rank=rank)
            for rank, user in enumerate(
                db
                .query(UserModel)
                .order_by(UserModel.points.desc())
                .where(UserModel.role != Role.ADMIN),
                1,
            )
        ],
    )


def compute_rank(db: Session, user_id: UUID) -> int:
    subq = (
        db
        .query(
            UserModel,
            func.row_number().over(order_by=UserModel.points.desc()).label("rownum"),
        )
        .where(UserModel.role != Role.ADMIN)
        .subquery()
    )

    rank: Sequence[int] = db.query(subq.c.rownum).where(subq.c.id == user_id).first()

    if not rank:
        return 0

    return rank[0]


@router.get(
    "/results",
    responses={
        status.HTTP_401_UNAUTHORIZED: {"model": ErrorOut},
        status.HTTP_422_UNPROCESSABLE_CONTENT: {"model": ValidationErrorOut},
    },
)
def retrieve_user_results(
    user: Annotated[UserModel, Depends(require_user)],
    db: Annotated[Session, Depends(get_db)],
) -> GenericOut[UserResult]:
    rank = compute_rank(db, user.id)

    user_result = UserResult.from_instance(user, rank=rank)

    return GenericOut(result=user_result)



"""Remove refresh token table

Revision ID: ce42052974e9
Revises: e5f9dac72943
Create Date: 2025-08-04 00:05:58.830918

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "ce42052974e9"
down_revision = "e5f9dac72943"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("refresh_token")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "refresh_token",
        sa.Column("id", sa.UUID(), autoincrement=False, nullable=False),
        sa.Column("user_id", sa.UUID(), autoincrement=False, nullable=False),
        sa.Column(
            "expiration",
            postgresql.TIMESTAMP(timezone=True),
            autoincrement=False,
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
            name=op.f("refresh_token_user_id_fkey"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("refresh_token_pkey")),
    )
    # ### end Alembic commands ###



import contextlib
from enum import Enum
from typing import TYPE_CHECKING, Optional
from uuid import UUID, uuid4

import sqlalchemy as sa
from argon2 import PasswordHasher
from argon2.exceptions import VerificationError
from sqlalchemy import CheckConstraint, UniqueConstraint
from sqlalchemy import Enum as SqlEnum
from sqlalchemy.dialects.postgresql import UUID as DB_UUID
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship

if TYPE_CHECKING:
    from sqlalchemy.orm import Session

ph = PasswordHasher()


class Base(DeclarativeBase):
    pass


class Role(Enum):
    USER = "user"
    ADMIN = "admin"


class UserModel(Base):
    __tablename__ = "user"
    id: Mapped[UUID] = mapped_column(DB_UUID(), primary_key=True, nullable=False, default=uuid4)
    name: Mapped[str] = mapped_column(sa.String(100), unique=True, nullable=False)
    first_name: Mapped[str] = mapped_column(sa.String(100), nullable=False)
    last_name: Mapped[str] = mapped_column(sa.String(100), nullable=False)
    password: Mapped[str] = mapped_column(sa.String(100), nullable=False)
    role: Mapped[Role] = mapped_column(SqlEnum(Role), nullable=False)
    number_match_guess: Mapped[int] = mapped_column(
        sa.Integer,
        CheckConstraint("number_match_guess>=0"),
        nullable=False,
        default=0,
    )
    number_score_guess: Mapped[int] = mapped_column(
        sa.Integer,
        CheckConstraint("number_score_guess>=0"),
        nullable=False,
        default=0,
    )
    number_qualified_teams_guess: Mapped[int] = mapped_column(
        sa.Integer,
        CheckConstraint("number_qualified_teams_guess>=0"),
        nullable=False,
        default=0,
    )
    number_first_qualified_guess: Mapped[int] = mapped_column(
        sa.Integer,
        CheckConstraint("number_first_qualified_guess>=0"),
        nullable=False,
        default=0,
    )
    number_quarter_final_guess: Mapped[int] = mapped_column(
        sa.Integer,
        CheckConstraint("number_quarter_final_guess>=0"),
        nullable=False,
        default=0,
    )
    number_semi_final_guess: Mapped[int] = mapped_column(
        sa.Integer,
        CheckConstraint("number_semi_final_guess>=0"),
        nullable=False,
        default=0,
    )
    number_final_guess: Mapped[int] = mapped_column(
        sa.Integer,
        CheckConstraint("number_final_guess>=0"),
        nullable=False,
        default=0,
    )
    number_winner_guess: Mapped[int] = mapped_column(
        sa.Integer,
        CheckConstraint("number_winner_guess>=0"),
        nullable=False,
        default=0,
    )

    points: Mapped[float] = mapped_column(
        sa.Float,
        CheckConstraint("points>=0"),
        nullable=False,
        default=0,
    )

    matches: Mapped[list["MatchModel"]] = relationship(
        "MatchModel",
        back_populates="user",
        lazy="dynamic",
        cascade="all, delete",
        passive_deletes=True,
    )

    def __init__(
        self,
        name: str,
        first_name: str,
        last_name: str,
        password: str,
        role: Role,
    ) -> None:
        self.name = name
        self.first_name = first_name
        self.last_name = last_name
        self.password = ph.hash(password)
        self.role = role

    @hybrid_property
    def full_name(self) -> str:
        return f"{self.first_name} {self.last_name}"

    @classmethod
    def authenticate(cls, db: "Session", name: str, password: str) -> Optional["UserModel"]:
        user = db.query(cls).filter_by(name=name).first()

        # user not found, verify password with itself to avoid timing attack
        if user is None:
            with contextlib.suppress(VerificationError):
                ph.verify(ph.hash(password), password)

            return None

        # user found
        try:
            ph.verify(user.password, password)
        except VerificationError:
            return None

        return user

    def change_password(self, new_password: str) -> None:
        self.password = ph.hash(new_password)


class ScoreBetModel(Base):
    __tablename__ = "score_bet"

    id: Mapped[UUID] = mapped_column(DB_UUID(), primary_key=True, nullable=False, default=uuid4)

    match_id: Mapped[UUID] = mapped_column(
        DB_UUID(),
        sa.ForeignKey("match.id", ondelete="CASCADE"),
        nullable=False,
    )
    match: Mapped["MatchModel"] = relationship(
        "MatchModel",
        back_populates="score_bets",
        lazy="raise",
    )

    score1: Mapped[int | None] = mapped_column(
        sa.Integer,
        CheckConstraint("score1>=0"),
        default=None,
    )
    score2: Mapped[int | None] = mapped_column(
        sa.Integer,
        CheckConstraint("score2>=0"),
        default=None,
    )

    def is_invalid(self) -> bool:
        return None in {self.score1, self.score2}

    def is_valid(self) -> bool:
        return not self.is_invalid()

    def is_1_win(self) -> bool:
        return self.score1 is not None and self.score2 is not None and self.score1 > self.score2

    def is_draw(self) -> bool:
        return self.score1 is not None and self.score2 is not None and self.score1 == self.score2

    def is_2_win(self) -> bool:
        return self.score1 is not None and self.score2 is not None and self.score1 < self.score2

    def is_same_results(self, other: "ScoreBetModel") -> bool:
        return (
            (self.is_1_win() and other.is_1_win())
            or (self.is_draw() and other.is_draw())
            or (self.is_2_win() and other.is_2_win())
        )

    def is_same_scores(self, other: "ScoreBetModel") -> bool:
        return (
            self.is_valid()
            and other.is_valid()
            and self.score1 == other.score1
            and self.score2 == other.score2
        )


class BinaryBetModel(Base):
    __tablename__ = "binary_bet"
    id = mapped_column(DB_UUID(), primary_key=True, nullable=False, default=uuid4)

    match_id: Mapped[UUID] = mapped_column(
        DB_UUID(),
        sa.ForeignKey("match.id", ondelete="CASCADE"),
        nullable=False,
    )
    match: Mapped["MatchModel"] = relationship(
        "MatchModel",
        back_populates="binary_bets",
        lazy="raise",
    )

    is_one_won: Mapped[bool | None] = mapped_column(sa.Boolean, default=None)

    def bet_from_is_one_won(self) -> tuple[None, None] | tuple[bool, bool]:
        if self.is_one_won is None:
            return (None, None)

        return (self.is_one_won, not self.is_one_won)


class BetMapping(Enum):
    SCORE_BET = ScoreBetModel
    BINARY_BET = BinaryBetModel


class MatchReferenceModel(Base):
    __tablename__ = "match_reference"
    id: Mapped[UUID] = mapped_column(DB_UUID(), primary_key=True, nullable=False, default=uuid4)

    group_id: Mapped[UUID] = mapped_column(DB_UUID(), sa.ForeignKey("group.id"), nullable=False)

    index: Mapped[int] = mapped_column(sa.Integer, nullable=False)

    team1_id: Mapped[UUID] = mapped_column(DB_UUID(), sa.ForeignKey("team.id"), nullable=True)
    team2_id: Mapped[UUID] = mapped_column(DB_UUID(), sa.ForeignKey("team.id"), nullable=True)

    bet_type_from_match: Mapped[BetMapping] = mapped_column(SqlEnum(BetMapping), nullable=False)

    __table_args__ = (UniqueConstraint("group_id", "index", name="uq_columns_group_id_index"),)


class MatchModel(Base):
    __tablename__ = "match"
    id: Mapped[UUID] = mapped_column(DB_UUID(), primary_key=True, nullable=False, default=uuid4)

    group_id: Mapped[UUID] = mapped_column(DB_UUID(), sa.ForeignKey("group.id"), nullable=False)
    group: Mapped["GroupModel"] = relationship(
        "GroupModel",
        foreign_keys=group_id,
        backref="matches",
        lazy="raise",
    )

    index: Mapped[int] = mapped_column(sa.Integer, nullable=False)

    team1_id: Mapped[UUID | None] = mapped_column(
        DB_UUID(),
        sa.ForeignKey("team.id"),
        nullable=True,
    )
    team1: Mapped[Optional["TeamModel"]] = relationship(
        "TeamModel",
        foreign_keys=team1_id,
        lazy="raise",
    )

    team2_id: Mapped[UUID | None] = mapped_column(
        DB_UUID(),
        sa.ForeignKey("team.id"),
        nullable=True,
    )
    team2: Mapped[Optional["TeamModel"]] = relationship(
        "TeamModel",
        foreign_keys=team2_id,
        lazy="raise",
    )

    user_id: Mapped[UUID] = mapped_column(
        DB_UUID(),
        sa.ForeignKey("user.id", ondelete="CASCADE"),
        nullable=False,
    )
    user: Mapped[UserModel] = relationship("UserModel", foreign_keys=user_id, lazy="raise")

    score_bets: Mapped[list[ScoreBetModel]] = relationship(
        "ScoreBetModel",
        back_populates="match",
        lazy="dynamic",
        cascade="all, delete",
        passive_deletes=True,
    )
    binary_bets: Mapped[list[BinaryBetModel]] = relationship(
        "BinaryBetModel",
        back_populates="match",
        lazy="dynamic",
        cascade="all, delete",
        passive_deletes=True,
    )


class TeamModel(Base):
    __tablename__ = "team"
    id: Mapped[UUID] = mapped_column(DB_UUID(), primary_key=True, nullable=False, default=uuid4)
    code: Mapped[str] = mapped_column(sa.String(10), unique=True, nullable=False)
    description_fr: Mapped[str] = mapped_column(sa.String(100), unique=True, nullable=False)
    description_en: Mapped[str] = mapped_column(sa.String(100), unique=True, nullable=False)
    flag_url: Mapped[str] = mapped_column(sa.String(100), nullable=False)
    internal_flag_url: Mapped[str] = mapped_column(sa.String(300), nullable=False)


class GroupModel(Base):
    __tablename__ = "group"
    id: Mapped[UUID] = mapped_column(DB_UUID(), primary_key=True, nullable=False, default=uuid4)
    code: Mapped[str] = mapped_column(sa.String(2), unique=True, nullable=False)
    description_fr: Mapped[str] = mapped_column(sa.String(100), unique=True, nullable=False)
    description_en: Mapped[str] = mapped_column(sa.String(100), unique=True, nullable=False)
    index: Mapped[int] = mapped_column(sa.Integer, nullable=False)

    phase_id: Mapped[UUID] = mapped_column(DB_UUID(), sa.ForeignKey("phase.id"), nullable=False)
    phase: Mapped["PhaseModel"] = relationship("PhaseModel", backref="groups", lazy="raise")

    __table_args__ = (UniqueConstraint("index", name="uq_group_index"),)


class PhaseModel(Base):
    __tablename__ = "phase"
    id: Mapped[UUID] = mapped_column(DB_UUID(), primary_key=True, nullable=False, default=uuid4)
    code: Mapped[str] = mapped_column(sa.String(10), unique=True, nullable=False)
    description_fr: Mapped[str] = mapped_column(sa.String(100), nullable=False)
    description_en: Mapped[str] = mapped_column(sa.String(100), nullable=False)
    index: Mapped[int] = mapped_column(sa.Integer, nullable=False)

    __table_args__ = (UniqueConstraint("index", name="uq_phase_index"),)


class GroupPositionModel(Base):
    __tablename__ = "group_position"
    id: Mapped[UUID] = mapped_column(DB_UUID(), primary_key=True, nullable=False, default=uuid4)
    won: Mapped[int] = mapped_column(
        sa.Integer,
        CheckConstraint("won>=0"),
        nullable=False,
        default=0,
    )
    drawn: Mapped[int] = mapped_column(
        sa.Integer,
        CheckConstraint("drawn>=0"),
        nullable=False,
        default=0,
    )
    lost: Mapped[int] = mapped_column(
        sa.Integer,
        CheckConstraint("lost>=0"),
        nullable=False,
        default=0,
    )
    goals_for: Mapped[int] = mapped_column(
        sa.Integer,
        CheckConstraint("goals_for>=0"),
        nullable=False,
        default=0,
    )
    goals_against: Mapped[int] = mapped_column(
        sa.Integer,
        CheckConstraint("goals_against>=0"),
        nullable=False,
        default=0,
    )

    need_recomputation = mapped_column(sa.Boolean, nullable=False, default=False)

    user_id: Mapped[UUID] = mapped_column(
        DB_UUID(),
        sa.ForeignKey("user.id", ondelete="CASCADE"),
        nullable=False,
    )

    team_id: Mapped[UUID] = mapped_column(
        DB_UUID(),
        sa.ForeignKey("team.id"),
        nullable=False,
    )
    team: Mapped[TeamModel] = relationship("TeamModel", lazy="raise")

    group_id: Mapped[UUID] = mapped_column(
        DB_UUID(),
        sa.ForeignKey("group.id"),
        nullable=False,
    )

    @hybrid_property
    def played(self) -> int:
        return self.won + self.drawn + self.lost

    @hybrid_property
    def goals_difference(self) -> int:
        return self.goals_for - self.goals_against

    @hybrid_property
    def points(self) -> int:
        return self.won * 3 + self.drawn



from importlib.metadata import version

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from pydantic_settings import BaseSettings, SettingsConfigDict

from . import health_check
from .helpers.logging_helpers import setup_logging
from .helpers.settings import CookieSettings
from .v1.helpers.errors import set_exception_handler
from .v1.routers import bets as bets_router
from .v1.routers import binary_bets as binary_bets_router
from .v1.routers import groups as groups_router
from .v1.routers import phases as phases_router
from .v1.routers import results as results_router
from .v1.routers import rules as rules_router
from .v1.routers import score_bets as score_bets_router
from .v1.routers import teams as teams_router
from .v1.routers import users as users_router

GLOBAL_ENDPOINT = "api"
VERSION1 = "v1"


class Config(BaseSettings):
    debug: bool = False

    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8", extra="allow")


def create_app() -> FastAPI:
    # Initialize fastapi application
    config = Config()

    app = FastAPI(
        debug=config.debug,
        docs_url=f"/{GLOBAL_ENDPOINT}/docs",
        redoc_url=f"/{GLOBAL_ENDPOINT}/redoc",
        openapi_url=f"/{GLOBAL_ENDPOINT}/openapi.json",
        version=version("yak-server"),
        title="Yak API",
        description="Yak API",
    )

    # Include health check router
    app.include_router(health_check.router, prefix=f"/{GLOBAL_ENDPOINT}")

    # Include all routers
    v1_prefix = f"/{GLOBAL_ENDPOINT}/{VERSION1}"

    app.include_router(bets_router.router, prefix=v1_prefix)
    app.include_router(binary_bets_router.router, prefix=v1_prefix)
    app.include_router(groups_router.router, prefix=v1_prefix)
    app.include_router(phases_router.router, prefix=v1_prefix)
    app.include_router(results_router.router, prefix=v1_prefix)
    app.include_router(rules_router.router, prefix=v1_prefix)
    app.include_router(score_bets_router.router, prefix=v1_prefix)
    app.include_router(teams_router.router, prefix=v1_prefix)
    app.include_router(users_router.router, prefix=v1_prefix)

    # Set error handler
    set_exception_handler(app)

    # Set CORS
    cookie_settings = CookieSettings()

    app.add_middleware(
        CORSMiddleware,
        allow_origins=cookie_settings.allowed_origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    # Declare logger configuration for yak server
    setup_logging(debug=app.debug)

    return app



from typing import TYPE_CHECKING

import pytest

from testing.util import get_resources_path
from yak_server.cli.database import (
    MissingGroupDuringInitError,
    MissingPhaseDuringInitError,
    MissingTeamDuringInitError,
    delete_database,
    initialize_database,
)
from yak_server.database.models import (
    GroupModel,
    MatchReferenceModel,
    PhaseModel,
    TeamModel,
)
from yak_server.database.session import build_local_session_maker

if TYPE_CHECKING:
    from sqlalchemy import Engine


def test_missing_phase(engine_for_test: "Engine") -> None:
    with pytest.raises(MissingPhaseDuringInitError) as exception:
        initialize_database(engine_for_test, get_resources_path("test_missing_phase"))

    assert (
        str(exception.value) == "Error during database initialization: phase_code=GROUP not found."
    )


def test_missing_team1(engine_for_test: "Engine") -> None:
    with pytest.raises(MissingTeamDuringInitError) as exception:
        initialize_database(engine_for_test, get_resources_path("test_missing_team1"))

    assert str(exception.value) == "Error during database initialization: team_code=AD not found."


def test_missing_team2(engine_for_test: "Engine") -> None:
    with pytest.raises(MissingTeamDuringInitError) as exception:
        initialize_database(engine_for_test, get_resources_path("test_missing_team2"))

    assert str(exception.value) == "Error during database initialization: team_code=BR not found."


def test_missing_group(engine_for_test: "Engine") -> None:
    with pytest.raises(MissingGroupDuringInitError) as exception:
        initialize_database(engine_for_test, get_resources_path("test_missing_group"))

    assert str(exception.value) == "Error during database initialization: group_code=A not found."


def test_upsert(engine_for_test: "Engine") -> None:
    # First initialization: insert records
    delete_database(engine_for_test, debug=True)
    initialize_database(engine_for_test, get_resources_path("test_upsert"))

    local_session_maker = build_local_session_maker(engine_for_test)

    with local_session_maker() as db:
        phase = db.query(PhaseModel).filter_by(code="GROUP").one()
        assert phase.description_fr == "Phase de groupes"
        assert phase.description_en == "Group stage"
        assert phase.index == 1

        group = db.query(GroupModel).filter_by(code="A").one()
        assert group.description_fr == "Groupe A"
        assert group.description_en == "Group A"
        assert group.index == 1

        france = db.query(TeamModel).filter_by(code="FR").one()
        assert france.description_fr == "France"
        assert france.description_en == "France"

        brazil = db.query(TeamModel).filter_by(code="BR").one()
        assert brazil.description_fr == "Br\u00e9sil"
        assert brazil.description_en == "Brazil"

        match = db.query(MatchReferenceModel).filter_by(group_id=group.id, index=1).one()
        assert match.team1_id == france.id
        assert match.team2_id == brazil.id

    # Second initialization: upsert with updated data
    initialize_database(engine_for_test, get_resources_path("test_upsert_updated"))

    with local_session_maker() as db:
        # Phase should be updated
        phase = db.query(PhaseModel).filter_by(code="GROUP").one()
        assert phase.description_fr == "Phases de poules"
        assert phase.description_en == "Group phase"
        assert phase.index == 1

        # Group should be updated
        group = db.query(GroupModel).filter_by(code="A").one()
        assert group.description_fr == "Poule A"
        assert group.description_en == "Pool A"
        assert group.index == 1

        # Teams should be updated
        france = db.query(TeamModel).filter_by(code="FR").one()
        assert france.description_fr == "Les Bleus"
        assert france.description_en == "The Blues"

        brazil = db.query(TeamModel).filter_by(code="BR").one()
        assert brazil.description_fr == "Sele\u00e7\u00e3o"
        assert brazil.description_en == "The Selection"

        # Match teams should be swapped
        match = db.query(MatchReferenceModel).filter_by(group_id=group.id, index=1).one()
        assert match.team1_id == brazil.id
        assert match.team2_id == france.id

    # Cleanup
    delete_database(engine_for_test, debug=True)



from datetime import datetime
from functools import cache

from pydantic import AwareDatetime, DirectoryPath, HttpUrl
from pydantic_settings import BaseSettings, SettingsConfigDict

from .rules import Rules


class Settings(BaseSettings):
    competition: str
    data_folder: DirectoryPath
    rules: Rules
    official_results_url: HttpUrl

    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8", extra="allow")


@cache
def get_settings() -> Settings:
    return Settings()


class LockDatetimeSettings(BaseSettings):
    lock_datetime: AwareDatetime

    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8", extra="allow")


@cache
def get_lock_datetime() -> datetime:
    return LockDatetimeSettings().lock_datetime  # pragma: no cover


class AuthenticationSettings(BaseSettings):
    jwt_secret_key: str
    jwt_refresh_secret_key: str
    jwt_expiration_time: int
    jwt_refresh_expiration_time: int

    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8", extra="allow")


@cache
def get_authentication_settings() -> AuthenticationSettings:
    return AuthenticationSettings()  # pragma: no cover


class CookieSettings(BaseSettings):
    cookie_secure: bool = True
    cookie_domain: str = ""
    allowed_origins: list[str] = []

    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8", extra="allow")


@cache
def get_cookie_settings() -> CookieSettings:
    return CookieSettings()  # pragma: no cover



"""Initial migration.

Revision ID: f33dc66104ad
Revises:
Create Date: 2023-02-27 15:42:06.550841

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f33dc66104ad"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "phase",
        sa.Column("id", sa.String(length=100), nullable=False),
        sa.Column("code", sa.String(length=10), nullable=False),
        sa.Column("description", sa.String(length=100), nullable=False),
        sa.Column("index", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id", "code"),
        sa.UniqueConstraint("code"),
    )
    op.create_table(
        "team",
        sa.Column("id", sa.String(length=100), nullable=False),
        sa.Column("code", sa.String(length=10), nullable=False),
        sa.Column("description", sa.String(length=100), nullable=False),
        sa.Column("flag_url", sa.String(length=100), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("code"),
        sa.UniqueConstraint("description"),
    )
    op.create_table(
        "user",
        sa.Column("id", sa.String(length=100), nullable=False),
        sa.Column("name", sa.String(length=100), nullable=False),
        sa.Column("first_name", sa.String(length=100), nullable=False),
        sa.Column("last_name", sa.String(length=100), nullable=False),
        sa.Column("password", sa.String(length=100), nullable=False),
        sa.Column("number_match_guess", sa.Integer(), nullable=False),
        sa.Column("number_score_guess", sa.Integer(), nullable=False),
        sa.Column("number_qualified_teams_guess", sa.Integer(), nullable=False),
        sa.Column("number_first_qualified_guess", sa.Integer(), nullable=False),
        sa.Column("number_quarter_final_guess", sa.Integer(), nullable=False),
        sa.Column("number_semi_final_guess", sa.Integer(), nullable=False),
        sa.Column("number_final_guess", sa.Integer(), nullable=False),
        sa.Column("number_winner_guess", sa.Integer(), nullable=False),
        sa.Column("points", sa.Float(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name"),
    )
    op.create_table(
        "group",
        sa.Column("id", sa.String(length=100), nullable=False),
        sa.Column("code", sa.String(length=1), nullable=False),
        sa.Column("description", sa.String(length=100), nullable=False),
        sa.Column("index", sa.Integer(), nullable=False),
        sa.Column("phase_id", sa.String(length=100), nullable=False),
        sa.ForeignKeyConstraint(
            ["phase_id"],
            ["phase.id"],
        ),
        sa.PrimaryKeyConstraint("id", "code"),
        sa.UniqueConstraint("code"),
        sa.UniqueConstraint("description"),
    )
    op.create_table(
        "match",
        sa.Column("id", sa.String(length=100), nullable=False),
        sa.Column("group_id", sa.String(length=100), nullable=False),
        sa.Column("index", sa.Integer(), nullable=False),
        sa.Column("team1_id", sa.String(length=100), nullable=False),
        sa.Column("team2_id", sa.String(length=100), nullable=False),
        sa.ForeignKeyConstraint(
            ["group_id"],
            ["group.id"],
        ),
        sa.ForeignKeyConstraint(
            ["team1_id"],
            ["team.id"],
        ),
        sa.ForeignKeyConstraint(
            ["team2_id"],
            ["team.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "binary_bet",
        sa.Column("id", sa.String(length=100), nullable=False),
        sa.Column("user_id", sa.String(length=100), nullable=False),
        sa.Column("match_id", sa.String(length=100), nullable=False),
        sa.Column("locked", sa.Boolean(), nullable=True),
        sa.Column("is_one_won", sa.Boolean(), nullable=True),
        sa.ForeignKeyConstraint(
            ["match_id"],
            ["match.id"],
        ),
        sa.ForeignKeyConstraint(["user_id"], ["user.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "score_bet",
        sa.Column("id", sa.String(length=100), nullable=False),
        sa.Column("user_id", sa.String(length=100), nullable=False),
        sa.Column("match_id", sa.String(length=100), nullable=False),
        sa.Column("score1", sa.Integer(), nullable=True),
        sa.Column("score2", sa.Integer(), nullable=True),
        sa.Column("locked", sa.Boolean(), nullable=True),
        sa.ForeignKeyConstraint(
            ["match_id"],
            ["match.id"],
        ),
        sa.ForeignKeyConstraint(["user_id"], ["user.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("score_bet")
    op.drop_table("binary_bet")
    op.drop_table("match")
    op.drop_table("group")
    op.drop_table("user")
    op.drop_table("team")
    op.drop_table("phase")
    # ### end Alembic commands ###



"""Add role column to User

Revision ID: d09424392322
Revises: ce42052974e9
Create Date: 2025-08-16 15:39:29.687453

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d09424392322"
down_revision = "ce42052974e9"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    # Create the ENUM type first
    role_enum = sa.Enum("USER", "ADMIN", name="role")
    role_enum.create(op.get_bind())

    # Add the column as nullable first
    op.add_column("user", sa.Column("role", role_enum, nullable=True))

    # Update existing rows to have a default role
    op.execute("UPDATE \"user\" SET role = 'USER' WHERE role IS NULL")

    # Now make the column NOT NULL
    op.alter_column("user", "role", nullable=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    # Drop the column first
    op.drop_column("user", "role")

    # Then drop the ENUM type
    role_enum = sa.Enum("USER", "ADMIN", name="role")
    role_enum.drop(op.get_bind())
    # ### end Alembic commands ###



import json

import click

from yak_server import create_app
from yak_server.database.session import build_engine
from yak_server.helpers.settings import get_settings

from .admin import create_admin
from .database import create_database, delete_database, drop_database, initialize_database
from .env import init_env, write_app_env_file, write_db_env_file
from .migration import setup_migration
from .score_board import compute_score_board
from .sync import synchronize_official_results


def make_db_app() -> click.Group:
    @click.group()
    def db_app() -> None:
        """Database related commands"""

    @db_app.command()
    def create() -> None:
        """Create all database tables."""
        engine = build_engine()
        create_database(engine)

    @db_app.command()
    def init() -> None:
        """Initialize database."""
        engine = build_engine()
        settings = get_settings()
        initialize_database(engine, settings.data_folder)

    @db_app.command()
    def drop() -> None:
        """Drop all tables."""
        app = create_app()
        engine = build_engine()
        drop_database(engine, debug=app.debug)

    @db_app.command()
    def delete() -> None:
        """Delete all records."""
        app = create_app()
        engine = build_engine()
        delete_database(engine, debug=app.debug)

    @db_app.command()
    @click.option(
        "--password",
        prompt=True,
        hide_input=True,
        confirmation_prompt=True,
        required=True,
    )
    def admin(password: str) -> None:
        """Create admin account in database."""
        engine = build_engine()
        create_admin(password, engine)

    @db_app.command()
    @click.option(
        "-s",
        "--short",
        is_flag=True,
        default=False,
    )
    def migration(*, short: bool) -> None:
        """Help to run database migration scripts."""
        setup_migration(short=short)

    @db_app.command()
    def score_board() -> None:
        """Compute score board."""
        engine = build_engine()
        compute_score_board(engine)

    @db_app.command()
    def sync() -> None:
        """Synchronize official results and push them to admin with web
        scraping the world cup wikipedia page"""
        engine = build_engine()
        settings = get_settings()
        synchronize_official_results(engine, settings.official_results_url)

    return db_app


def make_env_typer() -> click.Group:
    @click.group()
    def env_app() -> None:
        """Environment related commands"""

    @env_app.command()
    @click.option("--debug/--no-debug", prompt=True, is_flag=True, default=False, show_default=True)
    @click.option(
        "-h",
        "--host",
        prompt=True,
        default="localhost",
        show_default=True,
        help="Database host",
    )
    @click.option("-u", "--user", prompt=True, help="Database username", required=True)
    @click.option("--password", prompt=True, hide_input=True, required=True)
    @click.option(
        "-p",
        "--port",
        prompt=True,
        default=5432,
        show_default=1,
        help="Database port",
        required=True,
    )
    @click.option(
        "--database",
        "-D",
        prompt=True,
        default="yak_db",
        help="Database name",
        show_default=True,
    )
    @click.option(
        "--jwt-expiration",
        "-j",
        prompt=True,
        default=3600,
        show_default=True,
        help="JWT expiration time in seconds",
    )
    @click.option(
        "--jwt-refresh-expiration",
        "-jrf",
        prompt=True,
        default=3600,
        show_default=True,
        help="JWT refresh expiration time in seconds",
    )
    @click.option("--competition", "-c", prompt=True, help="Competition name", required=True)
    def all(  # noqa: A001, PLR0913
        *,
        debug: bool,
        host: str,
        user: str,
        password: str,
        port: int,
        database: str,
        jwt_expiration: int,
        jwt_refresh_expiration: int,
        competition: str,
    ) -> None:
        """Build the env files you need to start the server."""
        init_env(
            debug,
            host,
            user,
            password,
            competition,
            database,
            jwt_expiration,
            jwt_refresh_expiration,
            port,
        )

    @env_app.command()
    @click.option(
        "-h",
        "--host",
        prompt=True,
        default="localhost",
        show_default=True,
        help="Database host",
    )
    @click.option("-u", "--user", prompt=True, help="Database username", required=True)
    @click.option("--password", prompt=True, hide_input=True, required=True)
    @click.option(
        "-p",
        "--port",
        prompt=True,
        default=5432,
        show_default=1,
        help="Database port",
        required=True,
    )
    @click.option(
        "--database",
        "-D",
        prompt=True,
        default="yak_db",
        help="Database name",
        show_default=True,
    )
    def db(*, host: str, user: str, password: str, port: int, database: str) -> None:
        """Build the env files you need to setup database."""
        write_db_env_file(host, user, password, port, database)

    @env_app.command()
    @click.option("--debug/--no-debug", prompt=True, is_flag=True, default=False, show_default=True)
    @click.option(
        "--jwt-expiration",
        "-j",
        prompt=True,
        default=3600,
        show_default=True,
        help="JWT expiration time in seconds",
    )
    @click.option(
        "--jwt-refresh-expiration",
        "-jrf",
        prompt=True,
        default=3600,
        show_default=True,
        help="JWT refresh expiration time in seconds",
    )
    @click.option("--competition", "-c", prompt=True, help="Competition name", required=True)
    def app(
        *,
        debug: bool,
        jwt_expiration: int,
        jwt_refresh_expiration: int,
        competition: str,
    ) -> None:
        """Build the env files you need to setup application."""
        write_app_env_file(debug, jwt_expiration, jwt_refresh_expiration, competition)

    return env_app


@click.group()
def app() -> None:
    """CLI for yak application"""


@app.command()
def openapi() -> None:
    """Print the openapi.json file."""
    app = create_app()
    click.echo(json.dumps(app.openapi(), separators=(",", ":")))


app.add_command(make_db_app(), name="db")
app.add_command(make_env_typer(), name="env")



from typing import Generic, TypeVar

from pydantic import BaseModel, ConfigDict

Result = TypeVar("Result")


class GenericOut(BaseModel, Generic[Result]):
    ok: bool = True
    result: Result

    model_config = ConfigDict(extra="forbid")


class ErrorOut(BaseModel):
    ok: bool = False
    error_code: int
    description: str

    model_config = ConfigDict(extra="forbid")


class SingleValidationErrorOut(BaseModel):
    field: str
    error: str


class ValidationErrorOut(BaseModel):
    ok: bool = False
    error_code: int

    description: list[SingleValidationErrorOut]

    model_config = ConfigDict(extra="forbid")



"""Add unique constraint to match reference

Revision ID: eb38661b5a0d
Revises: 1719e9432d78
Create Date: 2025-07-14 22:29:38.502892

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "eb38661b5a0d"
down_revision = "1719e9432d78"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_unique_constraint(
        "uq_columns_group_id_index",
        "match_reference",
        ["group_id", "index"],
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("uq_columns_group_id_index", "match_reference", type_="unique")
    # ### end Alembic commands ###



from pydantic import BaseModel

from .binary_bets import BinaryBetOut, BinaryBetWithGroupIdOut
from .group_rank import GroupPositionOut
from .groups import GroupOut, GroupWithPhaseIdOut
from .phases import PhaseOut
from .score_bets import ScoreBetOut, ScoreBetWithGroupIdOut


class AllBetsResponse(BaseModel):
    phases: list[PhaseOut]
    groups: list[GroupWithPhaseIdOut]
    score_bets: list[ScoreBetWithGroupIdOut]
    binary_bets: list[BinaryBetWithGroupIdOut]


class BetsByPhaseCodeResponse(BaseModel):
    phase: PhaseOut
    groups: list[GroupOut]
    score_bets: list[ScoreBetWithGroupIdOut]
    binary_bets: list[BinaryBetWithGroupIdOut]


class BetsByGroupCodeResponse(BaseModel):
    phase: PhaseOut
    group: GroupOut
    score_bets: list[ScoreBetOut]
    binary_bets: list[BinaryBetOut]


class GroupRankResponse(BaseModel):
    phase: PhaseOut
    group: GroupOut
    group_rank: list[GroupPositionOut]



import logging
from datetime import timedelta
from typing import Annotated

from fastapi import APIRouter, Depends, Response, status
from pydantic import UUID4
from sqlalchemy.orm import Session

from yak_server.database.models import Role, UserModel
from yak_server.helpers.authentication import (
    NameAlreadyExistsError,
    encode_bearer_token,
    signup_user,
)
from yak_server.helpers.cookies import clear_auth_cookies, set_auth_cookies
from yak_server.helpers.database import get_db
from yak_server.helpers.logging_helpers import (
    logged_in_successfully,
    modify_password_successfully,
    signed_up_successfully,
)
from yak_server.helpers.password_validator import (
    PasswordRequirements,
    PasswordRequirementsError,
)
from yak_server.helpers.settings import (
    AuthenticationSettings,
    CookieSettings,
    get_authentication_settings,
    get_cookie_settings,
)
from yak_server.v1.helpers.auth import (
    require_admin,
    require_refresh_token,
    require_user,
    user_from_token,
)
from yak_server.v1.helpers.errors import (
    InvalidCredentials,
    NameAlreadyExists,
    UnsatisfiedPasswordRequirements,
    UserNotFound,
)
from yak_server.v1.models.generic import ErrorOut, GenericOut, ValidationErrorOut
from yak_server.v1.models.users import (
    CurrentUserOut,
    LoginIn,
    LoginOut,
    ModifyUserIn,
    PasswordRequirementsOut,
    RefreshOut,
    SignupIn,
    SignupOut,
)

logger = logging.getLogger(__name__)

router = APIRouter(prefix="/users", tags=["users"])


@router.post(
    "/signup",
    status_code=status.HTTP_201_CREATED,
    responses={
        status.HTTP_400_BAD_REQUEST: {"model": ErrorOut},
        status.HTTP_409_CONFLICT: {"model": ErrorOut},
        status.HTTP_422_UNPROCESSABLE_CONTENT: {"model": ValidationErrorOut},
    },
)
def signup(
    signup_in: SignupIn,
    response: Response,
    db: Annotated[Session, Depends(get_db)],
    auth_settings: Annotated[AuthenticationSettings, Depends(get_authentication_settings)],
    cookie_settings: Annotated[CookieSettings, Depends(get_cookie_settings)],
) -> GenericOut[SignupOut]:
    try:
        user = signup_user(
            db,
            signup_in.name,
            signup_in.first_name,
            signup_in.last_name,
            signup_in.password,
            Role.USER,
        )
    except PasswordRequirementsError as password_requirements_error:
        raise UnsatisfiedPasswordRequirements(
            str(password_requirements_error),
        ) from password_requirements_error
    except NameAlreadyExistsError as name_already_exists_error:
        raise NameAlreadyExists(signup_in.name) from name_already_exists_error

    logger.info(signed_up_successfully(user.name))

    access_token = encode_bearer_token(
        sub=user.id,
        expiration_time=timedelta(seconds=auth_settings.jwt_expiration_time),
        secret_key=auth_settings.jwt_secret_key,
    )
    refresh_token = encode_bearer_token(
        sub=user.id,
        expiration_time=timedelta(seconds=auth_settings.jwt_refresh_expiration_time),
        secret_key=auth_settings.jwt_refresh_secret_key,
    )

    set_auth_cookies(
        response=response,
        access_token=access_token,
        access_expires_in=auth_settings.jwt_expiration_time,
        refresh_token=refresh_token,
        refresh_expires_in=auth_settings.jwt_refresh_expiration_time,
        settings=cookie_settings,
    )

    return GenericOut(
        result=SignupOut(
            id=user.id,
            name=user.name,
            access_token=access_token,
            access_expires_in=auth_settings.jwt_expiration_time,
            refresh_token=refresh_token,
            refresh_expires_in=auth_settings.jwt_refresh_expiration_time,
        ),
    )


@router.get(
    "/signup/password_requirements",
    responses={status.HTTP_422_UNPROCESSABLE_CONTENT: {"model": ValidationErrorOut}},
)
def password_requirements() -> GenericOut[PasswordRequirementsOut]:
    password_requirements = PasswordRequirements()

    return GenericOut(
        result=PasswordRequirementsOut(
            minimum_length=password_requirements.MINIMUM_LENGTH,
            uppercase=password_requirements.UPPERCASE,
            lowercase=password_requirements.LOWERCASE,
            digit=password_requirements.DIGIT,
            no_space=password_requirements.NO_SPACE,
        ),
    )


@router.post(
    "/login",
    status_code=status.HTTP_201_CREATED,
    responses={
        status.HTTP_401_UNAUTHORIZED: {"model": ErrorOut},
        status.HTTP_409_CONFLICT: {"model": ErrorOut},
        status.HTTP_422_UNPROCESSABLE_CONTENT: {"model": ValidationErrorOut},
    },
)
def login(
    login_in: LoginIn,
    response: Response,
    db: Annotated[Session, Depends(get_db)],
    auth_settings: Annotated[AuthenticationSettings, Depends(get_authentication_settings)],
    cookie_settings: Annotated[CookieSettings, Depends(get_cookie_settings)],
) -> GenericOut[LoginOut]:
    user = UserModel.authenticate(db, login_in.name, login_in.password)

    if not user:
        raise InvalidCredentials

    logger.info(logged_in_successfully(user.name))

    access_token = encode_bearer_token(
        sub=user.id,
        expiration_time=timedelta(seconds=auth_settings.jwt_expiration_time),
        secret_key=auth_settings.jwt_secret_key,
    )
    refresh_token = encode_bearer_token(
        sub=user.id,
        expiration_time=timedelta(seconds=auth_settings.jwt_refresh_expiration_time),
        secret_key=auth_settings.jwt_refresh_secret_key,
    )

    set_auth_cookies(
        response=response,
        access_token=access_token,
        access_expires_in=auth_settings.jwt_expiration_time,
        refresh_token=refresh_token,
        refresh_expires_in=auth_settings.jwt_refresh_expiration_time,
        settings=cookie_settings,
    )

    return GenericOut(
        result=LoginOut(
            id=user.id,
            name=user.name,
            access_token=access_token,
            access_expires_in=auth_settings.jwt_expiration_time,
            refresh_token=refresh_token,
            refresh_expires_in=auth_settings.jwt_refresh_expiration_time,
        ),
    )


@router.post(
    "/refresh",
    status_code=status.HTTP_201_CREATED,
    responses={
        status.HTTP_401_UNAUTHORIZED: {"model": ErrorOut},
        status.HTTP_422_UNPROCESSABLE_CONTENT: {"model": ValidationErrorOut},
    },
)
def refresh(
    response: Response,
    refresh_token_value: Annotated[str, Depends(require_refresh_token)],
    db: Annotated[Session, Depends(get_db)],
    auth_settings: Annotated[AuthenticationSettings, Depends(get_authentication_settings)],
    cookie_settings: Annotated[CookieSettings, Depends(get_cookie_settings)],
) -> GenericOut[RefreshOut]:
    user = user_from_token(db, auth_settings.jwt_refresh_secret_key, refresh_token_value)

    access_token = encode_bearer_token(
        sub=user.id,
        expiration_time=timedelta(seconds=auth_settings.jwt_expiration_time),
        secret_key=auth_settings.jwt_secret_key,
    )
    new_refresh_token = encode_bearer_token(
        sub=user.id,
        expiration_time=timedelta(seconds=auth_settings.jwt_refresh_expiration_time),
        secret_key=auth_settings.jwt_refresh_secret_key,
    )

    set_auth_cookies(
        response=response,
        access_token=access_token,
        access_expires_in=auth_settings.jwt_expiration_time,
        refresh_token=new_refresh_token,
        refresh_expires_in=auth_settings.jwt_refresh_expiration_time,
        settings=cookie_settings,
    )

    return GenericOut(
        result=RefreshOut(
            access_token=access_token,
            access_expires_in=auth_settings.jwt_expiration_time,
            refresh_token=new_refresh_token,
            refresh_expires_in=auth_settings.jwt_refresh_expiration_time,
        ),
    )


@router.patch(
    "/{user_id}",
    responses={
        status.HTTP_400_BAD_REQUEST: {"model": ErrorOut},
        status.HTTP_401_UNAUTHORIZED: {"model": ErrorOut},
        status.HTTP_404_NOT_FOUND: {"model": ErrorOut},
        status.HTTP_422_UNPROCESSABLE_CONTENT: {"model": ValidationErrorOut},
    },
)
def modify_user(
    user_id: UUID4,
    modify_user_in: ModifyUserIn,
    db: Annotated[Session, Depends(get_db)],
    _: Annotated[UserModel, Depends(require_admin)],
) -> GenericOut[CurrentUserOut]:
    user = db.query(UserModel).filter_by(id=user_id).first()

    if not user:
        raise UserNotFound(user_id)

    user.change_password(modify_user_in.password)

    db.commit()

    logger.info(modify_password_successfully(user.name))

    return GenericOut(result=CurrentUserOut.model_validate(user))


@router.get(
    "/current",
    responses={
        status.HTTP_401_UNAUTHORIZED: {"model": ErrorOut},
        status.HTTP_422_UNPROCESSABLE_CONTENT: {"model": ValidationErrorOut},
    },
)
def current_user(
    user: Annotated[UserModel, Depends(require_user)],
) -> GenericOut[CurrentUserOut]:
    return GenericOut(result=CurrentUserOut.model_validate(user))


@router.post(
    "/logout",
    responses={
        status.HTTP_422_UNPROCESSABLE_CONTENT: {"model": ValidationErrorOut},
    },
)
def logout(
    response: Response,
    cookie_settings: Annotated[CookieSettings, Depends(get_cookie_settings)],
) -> GenericOut[None]:
    clear_auth_cookies(response, cookie_settings)
    return GenericOut(result=None)



import logging
from datetime import datetime
from typing import Annotated

from fastapi import APIRouter, Depends, status
from pydantic import UUID4
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session, selectinload

from yak_server.database.models import GroupModel, MatchModel, ScoreBetModel, UserModel
from yak_server.helpers.bet_locking import is_locked
from yak_server.helpers.database import get_db
from yak_server.helpers.group_position import set_recomputation_flag
from yak_server.helpers.language import DEFAULT_LANGUAGE, Lang, get_language_description
from yak_server.helpers.logging_helpers import modify_score_bet_successfully
from yak_server.helpers.settings import get_lock_datetime
from yak_server.v1.helpers.auth import require_user
from yak_server.v1.helpers.errors import BetNotFound, LockedScoreBet, TeamNotFound
from yak_server.v1.models.generic import ErrorOut, GenericOut, ValidationErrorOut
from yak_server.v1.models.groups import GroupOut
from yak_server.v1.models.phases import PhaseOut
from yak_server.v1.models.score_bets import ModifyScoreBetIn, ScoreBetOut, ScoreBetResponse
from yak_server.v1.models.teams import FlagOut, TeamWithScoreOut

logger = logging.getLogger(__name__)

router = APIRouter(prefix="/score_bets", tags=["score_bets"])


def send_response(
    score_bet: ScoreBetModel,
    *,
    locked: bool,
    lang: Lang,
) -> GenericOut[ScoreBetResponse]:
    return GenericOut(
        result=ScoreBetResponse(
            phase=PhaseOut.from_instance(score_bet.match.group.phase, lang=lang),
            group=GroupOut.from_instance(score_bet.match.group, lang=lang),
            score_bet=ScoreBetOut(
                id=score_bet.id,
                locked=locked,
                team1=(
                    TeamWithScoreOut(
                        id=score_bet.match.team1.id,
                        code=score_bet.match.team1.code,
                        description=get_language_description(score_bet.match.team1, lang),
                        flag=FlagOut(url=score_bet.match.team1.internal_flag_url),
                        score=score_bet.score1,
                    )
                    if score_bet.match.team1 is not None
                    else None
                ),
                team2=(
                    TeamWithScoreOut(
                        id=score_bet.match.team2.id,
                        code=score_bet.match.team2.code,
                        description=get_language_description(score_bet.match.team2, lang),
                        flag=FlagOut(url=score_bet.match.team2.internal_flag_url),
                        score=score_bet.score2,
                    )
                    if score_bet.match.team2 is not None
                    else None
                ),
            ),
        ),
    )


@router.get(
    "/{bet_id}",
    responses={
        status.HTTP_401_UNAUTHORIZED: {"model": ErrorOut},
        status.HTTP_404_NOT_FOUND: {"model": ErrorOut},
        status.HTTP_422_UNPROCESSABLE_CONTENT: {"model": ValidationErrorOut},
    },
)
def retrieve_score_bet_by_id(
    bet_id: UUID4,
    db: Annotated[Session, Depends(get_db)],
    user: Annotated[UserModel, Depends(require_user)],
    lock_datetime: Annotated[datetime, Depends(get_lock_datetime)],
    lang: Lang = DEFAULT_LANGUAGE,
) -> GenericOut[ScoreBetResponse]:
    score_bet = (
        db
        .query(ScoreBetModel)
        .options(
            selectinload(ScoreBetModel.match)
            .selectinload(MatchModel.group)
            .selectinload(GroupModel.phase),
            selectinload(ScoreBetModel.match).selectinload(MatchModel.team1),
            selectinload(ScoreBetModel.match).selectinload(MatchModel.team2),
        )
        .join(ScoreBetModel.match)
        .where(MatchModel.user_id == user.id, ScoreBetModel.id == bet_id)
        .first()
    )

    if not score_bet:
        raise BetNotFound(bet_id)

    return send_response(score_bet, locked=is_locked(user, lock_datetime), lang=lang)


@router.patch(
    "/{bet_id}",
    responses={
        status.HTTP_401_UNAUTHORIZED: {"model": ErrorOut},
        status.HTTP_404_NOT_FOUND: {"model": ErrorOut},
        status.HTTP_422_UNPROCESSABLE_CONTENT: {"model": ValidationErrorOut},
    },
)
def modify_score_bet(
    bet_id: UUID4,
    modify_score_bet_in: ModifyScoreBetIn,
    db: Annotated[Session, Depends(get_db)],
    user: Annotated[UserModel, Depends(require_user)],
    lock_datetime: Annotated[datetime, Depends(get_lock_datetime)],
    lang: Lang = DEFAULT_LANGUAGE,
) -> GenericOut[ScoreBetResponse]:
    if is_locked(user, lock_datetime):
        raise LockedScoreBet

    score_bet = (
        db
        .query(ScoreBetModel)
        .options(
            selectinload(ScoreBetModel.match)
            .selectinload(MatchModel.group)
            .selectinload(GroupModel.phase),
            selectinload(ScoreBetModel.match).selectinload(MatchModel.team1),
            selectinload(ScoreBetModel.match).selectinload(MatchModel.team2),
        )
        .join(ScoreBetModel.match)
        .where(MatchModel.user_id == user.id, ScoreBetModel.id == bet_id)
        .with_for_update()
        .first()
    )

    if not score_bet:
        raise BetNotFound(bet_id)

    logger.info(
        modify_score_bet_successfully(
            user.name,
            score_bet,
            modify_score_bet_in.team1.score if modify_score_bet_in.team1 else None,
            modify_score_bet_in.team2.score if modify_score_bet_in.team2 else None,
        ),
    )

    if modify_score_bet_in.team1 is not None:
        if "id" in modify_score_bet_in.team1.model_fields_set:
            score_bet.match.team1_id = modify_score_bet_in.team1.id

            try:
                db.flush()
            except IntegrityError as integrity_error:
                db.rollback()
                # team1.id is not None due to: "id" in modify_score_bet_in.team1.model_fields_set
                # being true
                raise TeamNotFound(modify_score_bet_in.team1.id) from integrity_error  # type: ignore[arg-type]

        if "score" in modify_score_bet_in.team1.model_fields_set:
            score_bet.score1 = modify_score_bet_in.team1.score

    if modify_score_bet_in.team2 is not None:
        if "id" in modify_score_bet_in.team2.model_fields_set:
            score_bet.match.team2_id = modify_score_bet_in.team2.id

            try:
                db.flush()
            except IntegrityError as integrity_error:
                db.rollback()
                # team2.id is not None due to: "id" in modify_score_bet_in.team2.model_fields_set
                # being true
                raise TeamNotFound(modify_score_bet_in.team2.id) from integrity_error  # type: ignore[arg-type]

        if "score" in modify_score_bet_in.team2.model_fields_set:
            score_bet.score2 = modify_score_bet_in.team2.score

    set_recomputation_flag(db, score_bet.match.team1_id, user.id)
    set_recomputation_flag(db, score_bet.match.team2_id, user.id)

    db.commit()
    db.refresh(score_bet)

    return send_response(score_bet, locked=is_locked(user, lock_datetime), lang=lang)



from collections.abc import Generator
from functools import cache

from sqlalchemy.orm import Session, sessionmaker

from yak_server.database.session import build_engine, build_local_session_maker


@cache
def _get_sqlalchemy_session_maker() -> sessionmaker[Session]:
    engine = build_engine()
    return build_local_session_maker(engine)


def get_db() -> Generator[Session, None, None]:
    local_session_maker = _get_sqlalchemy_session_maker()

    with local_session_maker() as db:
        yield db



from typing import TYPE_CHECKING

from yak_server.database.models import UserModel
from yak_server.database.session import build_local_session_maker
from yak_server.helpers.rules.compute_points import compute_points
from yak_server.helpers.settings import get_settings
from yak_server.v1.helpers.errors import NoAdminUser

if TYPE_CHECKING:
    from sqlalchemy import Engine


class ComputePointsRuleNotDefinedError(Exception):
    def __init__(self) -> None:
        super().__init__("Compute points rule is not defined.")


def compute_score_board(engine: "Engine") -> None:
    local_session_maker = build_local_session_maker(engine)

    with local_session_maker() as db:
        admin = db.query(UserModel).filter_by(name="admin").first()

        if admin is None:
            raise NoAdminUser

        rule_config = get_settings().rules.compute_points

        if rule_config is None:
            raise ComputePointsRuleNotDefinedError

        compute_points(db, admin, rule_config)



import logging
from datetime import datetime
from typing import Annotated

from fastapi import APIRouter, Depends, status
from pydantic import UUID4
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session, selectinload

from yak_server.database.models import BinaryBetModel, GroupModel, MatchModel, UserModel
from yak_server.helpers.bet_locking import is_locked
from yak_server.helpers.database import get_db
from yak_server.helpers.language import DEFAULT_LANGUAGE, Lang, get_language_description
from yak_server.helpers.logging_helpers import modify_binary_bet_successfully
from yak_server.helpers.settings import get_lock_datetime
from yak_server.v1.helpers.auth import require_user
from yak_server.v1.helpers.errors import BetNotFound, LockedBinaryBet, TeamNotFound
from yak_server.v1.models.binary_bets import BinaryBetOut, BinaryBetResponse, ModifyBinaryBetIn
from yak_server.v1.models.generic import ErrorOut, GenericOut, ValidationErrorOut
from yak_server.v1.models.groups import GroupOut
from yak_server.v1.models.phases import PhaseOut
from yak_server.v1.models.teams import FlagOut, TeamWithWonOut

logger = logging.getLogger(__name__)

router = APIRouter(prefix="/binary_bets", tags=["binary_bets"])


def send_response(
    binary_bet: BinaryBetModel,
    *,
    locked: bool,
    lang: Lang,
) -> GenericOut[BinaryBetResponse]:
    return GenericOut(
        result=BinaryBetResponse(
            phase=PhaseOut.from_instance(binary_bet.match.group.phase, lang=lang),
            group=GroupOut.from_instance(binary_bet.match.group, lang=lang),
            binary_bet=BinaryBetOut(
                id=binary_bet.id,
                locked=locked,
                team1=(
                    TeamWithWonOut(
                        id=binary_bet.match.team1.id,
                        code=binary_bet.match.team1.code,
                        description=get_language_description(binary_bet.match.team1, lang),
                        flag=FlagOut(url=binary_bet.match.team1.internal_flag_url),
                        won=binary_bet.bet_from_is_one_won()[0],
                    )
                    if binary_bet.match.team1 is not None
                    else None
                ),
                team2=(
                    TeamWithWonOut(
                        id=binary_bet.match.team2.id,
                        code=binary_bet.match.team2.code,
                        description=get_language_description(binary_bet.match.team2, lang),
                        flag=FlagOut(url=binary_bet.match.team2.internal_flag_url),
                        won=binary_bet.bet_from_is_one_won()[1],
                    )
                    if binary_bet.match.team2 is not None
                    else None
                ),
            ),
        ),
    )


@router.get(
    "/{bet_id}",
    responses={
        status.HTTP_401_UNAUTHORIZED: {"model": ErrorOut},
        status.HTTP_404_NOT_FOUND: {"model": ErrorOut},
        status.HTTP_422_UNPROCESSABLE_CONTENT: {"model": ValidationErrorOut},
    },
)
def retrieve_binary_bet_by_id(
    bet_id: UUID4,
    db: Annotated[Session, Depends(get_db)],
    user: Annotated[UserModel, Depends(require_user)],
    lock_datetime: Annotated[datetime, Depends(get_lock_datetime)],
    lang: Lang = DEFAULT_LANGUAGE,
) -> GenericOut[BinaryBetResponse]:
    binary_bet = (
        db
        .query(BinaryBetModel)
        .options(
            selectinload(BinaryBetModel.match)
            .selectinload(MatchModel.group)
            .selectinload(GroupModel.phase),
            selectinload(BinaryBetModel.match).selectinload(MatchModel.group),
            selectinload(BinaryBetModel.match).selectinload(MatchModel.team1),
            selectinload(BinaryBetModel.match).selectinload(MatchModel.team2),
        )
        .join(BinaryBetModel.match)
        .where(MatchModel.user_id == user.id, BinaryBetModel.id == bet_id)
        .first()
    )

    if not binary_bet:
        raise BetNotFound(bet_id)

    return send_response(binary_bet, locked=is_locked(user, lock_datetime), lang=lang)


@router.patch(
    "/{bet_id}",
    responses={
        status.HTTP_401_UNAUTHORIZED: {"model": ErrorOut},
        status.HTTP_404_NOT_FOUND: {"model": ErrorOut},
        status.HTTP_422_UNPROCESSABLE_CONTENT: {"model": ValidationErrorOut},
    },
)
def modify_binary_bet_by_id(
    bet_id: UUID4,
    modify_binary_bet_in: ModifyBinaryBetIn,
    db: Annotated[Session, Depends(get_db)],
    user: Annotated[UserModel, Depends(require_user)],
    lock_datetime: Annotated[datetime, Depends(get_lock_datetime)],
    lang: Lang = DEFAULT_LANGUAGE,
) -> GenericOut[BinaryBetResponse]:
    if is_locked(user, lock_datetime):
        raise LockedBinaryBet

    binary_bet = (
        db
        .query(BinaryBetModel)
        .options(
            selectinload(BinaryBetModel.match)
            .selectinload(MatchModel.group)
            .selectinload(GroupModel.phase),
            selectinload(BinaryBetModel.match).selectinload(MatchModel.group),
            selectinload(BinaryBetModel.match).selectinload(MatchModel.team1),
            selectinload(BinaryBetModel.match).selectinload(MatchModel.team2),
        )
        .join(BinaryBetModel.match)
        .where(MatchModel.user_id == user.id, BinaryBetModel.id == bet_id)
        .first()
    )

    if not binary_bet:
        raise BetNotFound(bet_id)

    logger.info(
        modify_binary_bet_successfully(
            user.name,
            binary_bet,
            new_is_one_won=modify_binary_bet_in.is_one_won,
        ),
    )

    if "is_one_won" in modify_binary_bet_in.model_fields_set:
        binary_bet.is_one_won = modify_binary_bet_in.is_one_won

    if (
        modify_binary_bet_in.team1 is not None
        and "id" in modify_binary_bet_in.team1.model_fields_set
    ):
        binary_bet.match.team1_id = modify_binary_bet_in.team1.id

        try:
            db.flush()
        except IntegrityError as integrity_error:
            db.rollback()
            # team1.id is not None due to: "id" in modify_binary_bet_in.team1.model_fields_set
            # being true
            raise TeamNotFound(modify_binary_bet_in.team1.id) from integrity_error  # type: ignore[arg-type]

    if (
        modify_binary_bet_in.team2 is not None
        and "id" in modify_binary_bet_in.team2.model_fields_set
    ):
        binary_bet.match.team2_id = modify_binary_bet_in.team2.id

        try:
            db.flush()
        except IntegrityError as integrity_error:
            db.rollback()
            # team2.id is not None due to: "id" in modify_binary_bet_in.team2.model_fields_set
            # being true
            raise TeamNotFound(modify_binary_bet_in.team2.id) from integrity_error  # type: ignore[arg-type]

    db.commit()
    db.refresh(binary_bet)

    return send_response(binary_bet, locked=is_locked(user, lock_datetime), lang=lang)



import json
import os
from datetime import datetime, timezone
from http import HTTPStatus
from importlib.metadata import version
from pathlib import Path
from typing import TYPE_CHECKING

from click.testing import CliRunner
from starlette.testclient import TestClient

from testing.util import get_random_string
from yak_server.cli import app

if TYPE_CHECKING:
    import pytest
    from fastapi import FastAPI

runner = CliRunner()


def test_cli(app_with_valid_jwt_config: "FastAPI") -> None:
    # Check database drop
    result = runner.invoke(app, ["db", "drop"], env={"DEBUG": "1"})

    assert result.exit_code == 0

    # Check database creation
    result = runner.invoke(app, ["db", "create"])

    assert result.exit_code == 0

    # Check database initialization
    data_folder = str(Path(__file__).parents[1] / "yak_server" / "data" / "world_cup_2022")

    result = runner.invoke(
        app,
        ["db", "init"],
        env={
            "JWT_EXPIRATION_TIME": "1800",
            "JWT_REFRESH_EXPIRATION_TIME": "1800",
            "JWT_SECRET_KEY": get_random_string(128),
            "COMPETITION": "world_cup_2022",
            "DATA_FOLDER": data_folder,
            "LOCK_DATETIME": str(datetime.now(timezone.utc)),
            "RULES": "{}",
            "OFFICIAL_RESULTS_URL": "https://en.wikipedia.org/wiki/2022_FIFA_World_Cup",
        },
    )

    assert result.exit_code == 0

    # Check admin account creation
    admin_password = get_random_string(9)

    result = runner.invoke(
        app,
        ["db", "admin"],
        env={
            "JWT_EXPIRATION_TIME": "1800",
            "JWT_SECRET_KEY": get_random_string(128),
            "COMPETITION": "world_cup_2022",
            "DATA_FOLDER": data_folder,
            "LOCK_DATETIME": str(datetime.now(timezone.utc)),
            "RULES": "{}",
        },
        input=f"{admin_password}\n{admin_password}\n",
    )

    assert result.exit_code == 0

    client = TestClient(app_with_valid_jwt_config)

    response_login = client.post(
        "/api/v1/users/login",
        json={
            "name": "admin",
            "password": admin_password,
        },
    )

    assert response_login.status_code == HTTPStatus.CREATED

    auth_token = response_login.json()["result"]["access_token"]
    user_id = response_login.json()["result"]["id"]

    # Check records deletion
    result = runner.invoke(app, ["db", "delete"], env={**os.environ, "DEBUG": "1"})

    assert result.exit_code == 0

    # Check user cannot access after records are cleaned
    response_login_user_not_found = client.get(
        "/api/v1/bets",
        headers={"Authorization": f"Bearer {auth_token}"},
    )

    assert response_login_user_not_found.status_code == HTTPStatus.NOT_FOUND
    assert response_login_user_not_found.json() == {
        "ok": False,
        "error_code": HTTPStatus.NOT_FOUND,
        "description": f"User not found: {user_id}",
    }


def test_cli_admin_already_exists() -> None:
    # Check admin account creation
    admin_password = get_random_string(9)

    result = runner.invoke(
        app,
        ["db", "admin"],
        input=f"{admin_password}\n{admin_password}\n",
    )

    admin_already_exists_message = "Admin already exists"

    assert result.exit_code == 0
    assert admin_already_exists_message not in result.stdout

    result = runner.invoke(
        app,
        ["db", "admin"],
        input=f"{admin_password}\n{admin_password}\n",
    )

    assert result.exit_code == 0
    assert admin_already_exists_message in result.stdout


def test_db_migration_cli_with_alembic_present() -> None:
    result = runner.invoke(app, ["db", "migration"])

    assert result.exit_code == 0


def test_db_migration_cli_short_option() -> None:
    result = runner.invoke(app, ["db", "migration", "--short"])

    alembic_ini_path = (Path(__file__).parents[1] / "alembic.ini").resolve()

    assert result.exit_code == 0
    assert result.output == f"export ALEMBIC_CONFIG={alembic_ini_path}\n"

    result_with_short_name = result = runner.invoke(app, ["db", "migration", "-s"])

    assert result_with_short_name.exit_code == result.exit_code
    assert result_with_short_name.output == result.output


def test_db_migration_second_path(monkeypatch: "pytest.MonkeyPatch") -> None:
    """
    .
    └── yak-server
        └── yak_server
            └── alembic.ini
    """

    with runner.isolated_filesystem():
        yak_server_module = (Path.cwd() / "yak-server" / "yak_server").resolve()

        python_file_path = yak_server_module / "cli" / "migration.py"
        alembic_ini_path = yak_server_module / "alembic.ini"

        alembic_ini_path.parent.mkdir(parents=True)
        alembic_ini_path.touch()

        monkeypatch.setattr("yak_server.cli.migration.__file__", str(python_file_path))

        result = runner.invoke(app, ["db", "migration", "-s"])

        assert result.exit_code == 0
        assert result.output == f"export ALEMBIC_CONFIG={alembic_ini_path}\n"


def test_db_migration_cli_with_alembic_missing(monkeypatch: "pytest.MonkeyPatch") -> None:
    monkeypatch.setattr("yak_server.cli.migration.alembic", None)

    result = runner.invoke(app, ["db", "migration"])

    assert result.exit_code == 0
    assert (
        "To enable migration using alembic, please run: uv pip install yak-server[db_migration]"
        in result.output
    )


def test_get_openapi_spec() -> None:
    """
    Test that the OpenAPI spec is generated correctly.
    """

    result = runner.invoke(app, ["openapi"])

    assert result.exit_code == 0

    openapi_spec = json.loads(result.output)

    assert openapi_spec["openapi"] == "3.1.0"
    assert openapi_spec["info"]["version"] == version("yak-server")
    assert "paths" in openapi_spec
    assert "components" in openapi_spec



from pathlib import Path

import click

try:
    import alembic
except ImportError:  # pragma: no cover
    # Very common pattern for optional dependency imports
    alembic = None  # type: ignore[assignment]


def print_export_command(alembic_ini_path: Path) -> None:
    click.echo(f"export ALEMBIC_CONFIG={alembic_ini_path}")


def setup_migration(*, short: bool = False) -> None:
    alembic_ini_path = (Path(__file__).parents[1] / "alembic.ini").resolve()

    if not alembic_ini_path.exists():
        alembic_ini_path = (Path(__file__).parents[2] / "alembic.ini").resolve()

    if short is True:
        print_export_command(alembic_ini_path)
    else:
        click.echo(
            "To be able to run the database migration scripts, "
            "you need to run the following command:",
        )
        print_export_command(alembic_ini_path)
        click.echo()
        click.echo(
            "Follow this link for more information: "
            "https://alembic.sqlalchemy.org/en/latest/tutorial.html#editing-the-ini-file",
        )

        if alembic is None:
            click.echo()
            click.echo(
                "To enable migration using alembic, please run: "
                "uv pip install yak-server[db_migration]",
            )



import sys
from typing import TYPE_CHECKING, Optional

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from datetime import datetime

from .util import get_resources_path

if TYPE_CHECKING:
    from yak_server.helpers.rules import Rules


class MockSettings:
    def __init__(
        self,
        *,
        data_folder_relative: str | None = None,
        rules: Optional["Rules"] = None,
        official_results_url: str | None = None,
    ) -> None:
        self.data_folder = (
            get_resources_path(data_folder_relative) if data_folder_relative is not None else None
        )

        self.rules = rules
        self.official_results_url = official_results_url

    def __call__(self) -> Self:
        return self


class MockLockDatetime:
    def __init__(self, lock_datetime: datetime) -> None:
        self.lock_datetime = lock_datetime

    def __call__(self) -> datetime:
        return self.lock_datetime


class MockAuthenticationSettings:
    def __init__(
        self,
        *,
        jwt_secret_key: str,
        jwt_refresh_secret_key: str,
        jwt_expiration_time: int,
        jwt_refresh_expiration_time: int,
    ) -> None:
        self.jwt_secret_key = jwt_secret_key
        self.jwt_refresh_secret_key = jwt_refresh_secret_key
        self.jwt_expiration_time = jwt_expiration_time
        self.jwt_refresh_expiration_time = jwt_refresh_expiration_time

    def __call__(self) -> Self:
        return self


class MockCookieSettings:
    def __init__(
        self,
        *,
        cookie_secure: bool = False,
        cookie_domain: str = "",
        allowed_origins: list[str] | None = None,
    ) -> None:
        self.cookie_secure = cookie_secure
        self.cookie_domain = cookie_domain
        self.allowed_origins = allowed_origins or []

    def __call__(self) -> Self:
        return self



from datetime import datetime, timedelta, timezone
from enum import Enum
from typing import Any
from uuid import UUID

import jwt
from sqlalchemy.orm import Session, selectinload

from yak_server.database.models import (
    MatchModel,
    MatchReferenceModel,
    Role,
    ScoreBetModel,
    UserModel,
)

from .errors import name_already_exists_message
from .group_position import create_group_position
from .password_validator import validate_password


def encode_bearer_token(
    sub: UUID,
    expiration_time: timedelta,
    secret_key: str,
) -> str:
    return jwt.encode(
        {
            "sub": str(sub),
            "nbf": datetime.now(timezone.utc) - timedelta(seconds=3),
            "exp": datetime.now(timezone.utc) + expiration_time,
        },
        secret_key,
        algorithm="HS512",
    )


def decode_bearer_token(token: str, secret_key: str) -> Any:  # noqa: ANN401
    return jwt.decode(token, secret_key, algorithms=["HS512"])


class NameAlreadyExistsError(Exception):
    def __init__(self, name: str) -> None:
        super().__init__(name_already_exists_message(name))


def signup_user(
    db: Session,
    name: str,
    first_name: str,
    last_name: str,
    password: str,
    role: Role,
) -> UserModel:
    # Check existing user in db
    existing_user = db.query(UserModel).filter_by(name=name).first()
    if existing_user:
        raise NameAlreadyExistsError(name)

    # Validate password
    validate_password(password)

    # Initialize user and integrate in db
    user = UserModel(name, first_name, last_name, password, role)
    db.add(user)
    db.flush()

    # Initialize matches and bets and integrate in db
    for match_reference in db.query(MatchReferenceModel).all():
        match = MatchModel(
            team1_id=match_reference.team1_id,
            team2_id=match_reference.team2_id,
            index=match_reference.index,
            group_id=match_reference.group_id,
            user_id=user.id,
        )
        db.add(match)
        db.flush()

        db.add(match_reference.bet_type_from_match.value(match_id=match.id))
        db.flush()

    # Create group position records
    db.add_all(
        create_group_position(
            db
            .query(ScoreBetModel)
            .options(selectinload(ScoreBetModel.match))
            .join(ScoreBetModel.match)
            .filter_by(user_id=user.id),
        ),
    )
    db.commit()

    return user


class Permission(Enum):
    """Define different permission levels"""

    USER = "user"  # Basic authenticated user
    ADMIN = "admin"  # Administrative access


def has_permission(user: UserModel, required_permission: Permission) -> bool:
    """Check if user has the required permission level.

    Args:
        user: The user to check permissions for
        required_permission: The permission level required

    Returns:
        True if user has the required permission, False otherwise
    """
    permission_hierarchy = {
        Permission.USER: [Role.USER, Role.ADMIN],
        Permission.ADMIN: [Role.ADMIN],
    }

    return user.role in permission_hierarchy[required_permission]



from http import HTTPStatus
from typing import TYPE_CHECKING
from unittest.mock import ANY

from starlette.testclient import TestClient

from testing.util import get_random_string

if TYPE_CHECKING:
    from fastapi import FastAPI


def test_valid_auth(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config)

    user_name = get_random_string(6)
    first_name = get_random_string(6)
    last_name = get_random_string(6)
    password = get_random_string(8)

    # signup test
    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": user_name,
            "first_name": first_name,
            "last_name": last_name,
            "password": password,
        },
    )
    assert response_signup.status_code == HTTPStatus.CREATED
    assert response_signup.json() == {
        "ok": True,
        "result": {
            "id": ANY,
            "name": user_name,
            "access_token": ANY,
            "access_expires_in": 100,
            "refresh_token": ANY,
            "refresh_expires_in": 200,
        },
    }

    # login test
    response_login = client.post(
        "/api/v1/users/login",
        json={"name": user_name, "password": password},
    )
    assert response_login.status_code == HTTPStatus.CREATED
    assert response_login.json() == {
        "ok": True,
        "result": {
            "id": ANY,
            "name": user_name,
            "access_token": ANY,
            "access_expires_in": 100,
            "refresh_token": ANY,
            "refresh_expires_in": 200,
        },
    }

    # current user tests
    response_current_user = client.get("/api/v1/users/current")
    assert response_current_user.status_code == HTTPStatus.OK
    assert response_current_user.json() == {
        "ok": True,
        "result": {"id": ANY, "name": user_name},
    }


def test_double_signup(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config)

    user_name = get_random_string(6)
    first_name = get_random_string(10)
    last_name = get_random_string(11)
    password = get_random_string(15)

    # signup test
    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": user_name,
            "first_name": first_name,
            "last_name": last_name,
            "password": password,
        },
    )
    assert response_signup.status_code == HTTPStatus.CREATED
    assert response_signup.json() == {
        "ok": True,
        "result": {
            "id": ANY,
            "name": user_name,
            "access_token": ANY,
            "access_expires_in": 100,
            "refresh_token": ANY,
            "refresh_expires_in": 200,
        },
    }

    # Try to signup with the same user name
    response_second_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": user_name,
            "first_name": get_random_string(10),
            "last_name": get_random_string(12),
            "password": get_random_string(8),
        },
    )
    assert response_second_signup.status_code == HTTPStatus.CONFLICT
    assert response_second_signup.json() == {
        "ok": False,
        "error_code": HTTPStatus.CONFLICT,
        "description": f"Name already exists: {user_name}",
    }


def test_login_wrong_name(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config)

    response_login = client.post(
        "/api/v1/users/login",
        json={
            "name": get_random_string(15),
            "password": get_random_string(12),
        },
    )

    assert response_login.status_code == HTTPStatus.UNAUTHORIZED
    assert response_login.json() == {
        "ok": False,
        "error_code": HTTPStatus.UNAUTHORIZED,
        "description": "Invalid credentials",
    }


def test_login_wrong_password(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config)

    user_name = get_random_string(6)

    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": user_name,
            "first_name": get_random_string(10),
            "last_name": get_random_string(11),
            "password": get_random_string(12),
        },
    )

    assert response_signup.status_code == HTTPStatus.CREATED

    response_login = client.post(
        "/api/v1/users/login",
        json={"name": user_name, "password": get_random_string(12)},
    )

    assert response_login.status_code == HTTPStatus.UNAUTHORIZED
    assert response_login.json() == {
        "ok": False,
        "error_code": HTTPStatus.UNAUTHORIZED,
        "description": "Invalid credentials",
    }


def test_invalid_token(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config)

    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": get_random_string(6),
            "first_name": get_random_string(2),
            "last_name": get_random_string(8),
            "password": get_random_string(8),
        },
    )

    auth_token = response_signup.json()["result"]["access_token"]

    response_get_all_bets_with_cropped_token = client.get(
        "/api/v1/bets",
        headers={"Authorization": f"Bearer {auth_token[:17]}{auth_token[12:]}"},
    )

    assert response_get_all_bets_with_cropped_token.status_code == HTTPStatus.UNAUTHORIZED
    assert response_get_all_bets_with_cropped_token.json() == {
        "ok": False,
        "error_code": HTTPStatus.UNAUTHORIZED,
        "description": "Invalid access token, authentication required",
    }


def test_expired_token(app_with_null_jwt_expiration_time: "FastAPI") -> None:
    client = TestClient(app_with_null_jwt_expiration_time)

    user_name = get_random_string(6)
    password = get_random_string(15)

    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": user_name,
            "first_name": get_random_string(2),
            "last_name": get_random_string(8),
            "password": password,
        },
    )

    assert response_signup.status_code == HTTPStatus.CREATED

    auth_token = response_signup.json()["result"]["access_token"]

    response_current_user = client.get(
        "/api/v1/users/current",
        headers={"Authorization": f"Bearer {auth_token}"},
    )

    assert response_current_user.status_code == HTTPStatus.UNAUTHORIZED
    assert response_current_user.json() == {
        "ok": False,
        "error_code": HTTPStatus.UNAUTHORIZED,
        "description": "Expired access token, re-authentication required",
    }


def test_invalid_signup_body(app_with_valid_jwt_config: "FastAPI") -> None:
    name = get_random_string(10)
    first_name = get_random_string(12)
    last_name = get_random_string(6)
    password = get_random_string(5)

    client = TestClient(app_with_valid_jwt_config)

    # Try to signup with invalid body
    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": name,
            "first_name": first_name,
            "last_name": last_name,
            "passwor": password,
        },
    )

    assert response_signup.status_code == HTTPStatus.UNPROCESSABLE_ENTITY
    assert response_signup.json() == {
        "ok": False,
        "error_code": HTTPStatus.UNPROCESSABLE_ENTITY,
        "description": [
            {"field": "body -> password", "error": "Field required"},
            {"field": "body -> passwor", "error": "Extra inputs are not permitted"},
        ],
    }


def test_invalid_login_body(app_with_valid_jwt_config: "FastAPI") -> None:
    user_name = get_random_string(6)
    password = get_random_string(10)

    client = TestClient(app_with_valid_jwt_config)

    response_login = client.post(
        "/api/v1/users/login",
        json={
            "nme": user_name,
            "password": password,
        },
    )

    assert response_login.status_code == HTTPStatus.UNPROCESSABLE_ENTITY
    assert response_login.json() == {
        "ok": False,
        "error_code": HTTPStatus.UNPROCESSABLE_ENTITY,
        "description": [
            {"field": "body -> name", "error": "Field required"},
            {"field": "body -> nme", "error": "Extra inputs are not permitted"},
        ],
    }


def test_no_token(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config, cookies={})

    response = client.get(
        "/api/v1/users/current",
    )

    assert response.status_code == HTTPStatus.UNAUTHORIZED
    assert response.json() == {
        "ok": False,
        "error_code": HTTPStatus.UNAUTHORIZED,
        "description": "Invalid access token, authentication required",
    }


def test_logout(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config)

    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": get_random_string(6),
            "first_name": get_random_string(2),
            "last_name": get_random_string(8),
            "password": get_random_string(8),
        },
    )

    assert response_signup.status_code == HTTPStatus.CREATED

    response_logout = client.post("/api/v1/users/logout")

    assert response_logout.status_code == HTTPStatus.OK
    assert response_logout.json() == {"ok": True, "result": None}

    # After logout, cookies should be cleared, so accessing a protected endpoint should fail
    response_current_user = client.get("/api/v1/users/current")

    assert response_current_user.status_code == HTTPStatus.UNAUTHORIZED
    assert response_current_user.json() == {
        "ok": False,
        "error_code": HTTPStatus.UNAUTHORIZED,
        "description": "Invalid access token, authentication required",
    }


def test_non_compliant_password(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config)

    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": get_random_string(2),
            "first_name": get_random_string(2),
            "last_name": get_random_string(8),
            "password": get_random_string(6),
        },
    )

    assert response_signup.status_code == HTTPStatus.BAD_REQUEST
    assert response_signup.json() == {
        "ok": False,
        "error_code": HTTPStatus.BAD_REQUEST,
        "description": (
            "Unsatisfied password requirements. Password is too short. Minimum size is 8."
        ),
    }



from http import HTTPStatus
from typing import TYPE_CHECKING
from uuid import uuid4

from starlette.testclient import TestClient

from testing.util import get_random_string
from yak_server.cli.admin import create_admin

if TYPE_CHECKING:
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_modify_password(app_with_valid_jwt_config: "FastAPI", engine_for_test: "Engine") -> None:
    client = TestClient(app_with_valid_jwt_config)

    password = get_random_string(9)
    other_user_name = get_random_string(6)

    # Create admin account
    create_admin(password, engine_for_test)

    response_login_admin = client.post(
        "/api/v1/users/login",
        json={
            "name": "admin",
            "password": password,
        },
    )

    authentication_token = response_login_admin.json()["result"]["access_token"]

    # Create non admin user account
    response_signup_glepape = client.post(
        "/api/v1/users/signup",
        json={
            "name": other_user_name,
            "first_name": "Guillaume",
            "last_name": "Le Pape",
            "password": get_random_string(18),
        },
    )

    user_id = response_signup_glepape.json()["result"]["id"]
    authentication_token_glepape = response_signup_glepape.json()["result"]["access_token"]

    # Check update is properly process
    new_password_other_user = get_random_string(15)

    response_modify_password = client.patch(
        f"/api/v1/users/{user_id}",
        headers={"Authorization": f"Bearer {authentication_token}"},
        json={"password": new_password_other_user},
    )

    assert response_modify_password.status_code == HTTPStatus.OK
    assert response_modify_password.json()["result"]["name"] == other_user_name

    # Check successful login with the new password
    response_login_glepape = client.post(
        "/api/v1/users/login",
        json={"name": other_user_name, "password": new_password_other_user},
    )

    assert response_login_glepape.status_code == HTTPStatus.CREATED
    assert response_login_glepape.json()["result"]["name"] == other_user_name

    # Check glepape user cannot update any password
    response_modify_password_with_glepape_user = client.patch(
        f"/api/v1/users/{user_id}",
        headers={"Authorization": f"Bearer {authentication_token_glepape}"},
        json={"password": "new_new_password"},
    )

    assert response_modify_password_with_glepape_user.status_code == HTTPStatus.UNAUTHORIZED
    assert (
        response_modify_password_with_glepape_user.json()["description"]
        == "Unauthorized access to admin API"
    )

    # Check update is rejected if body does not contain any password
    response_wrong_input = client.patch(
        f"/api/v1/users/{user_id}",
        headers={"Authorization": f"Bearer {authentication_token}"},
        json={"name": other_user_name},
    )

    assert response_wrong_input.status_code == HTTPStatus.UNPROCESSABLE_ENTITY
    assert response_wrong_input.json() == {
        "ok": False,
        "error_code": HTTPStatus.UNPROCESSABLE_ENTITY,
        "description": [
            {"field": "body -> password", "error": "Field required"},
            {"field": "body -> name", "error": "Extra inputs are not permitted"},
        ],
    }

    # Check call is rejected if user_id is invalid
    invalid_user_id = uuid4()

    response_wrong_input = client.patch(
        f"/api/v1/users/{invalid_user_id}",
        headers={"Authorization": f"Bearer {authentication_token}"},
        json={"password": "new_password_for_unknown_user"},
    )

    assert response_wrong_input.status_code == HTTPStatus.NOT_FOUND
    assert response_wrong_input.json()["error_code"] == HTTPStatus.NOT_FOUND
    assert response_wrong_input.json()["description"] == f"User not found: {invalid_user_id}"
